Add LoadPath "~/ba/casestudy".
Require Export stlc_num.
Require Export wn.

Lemma substitutivity_h {n} (e1 : tm (S n)) e2 t :  t = subst_tm (e2 .: var_tm) e1 ->  (app (lam e1) e2) ≻ t.
Proof.
  intros h. rewrite h. constructor. 
Qed.
  
Lemma substitutivity {n m} s t σ : s ≻ t -> subst_tm σ s ≻ @subst_tm n m σ t.
Proof.
  intros H. revert m σ. induction H; intros m σ; try ( asimpl; constructor; eauto). 
  - asimpl. apply substitutivity_h. asimpl. reflexivity.
Qed.

Lemma substitutivity_ren {n m} s t ξ : s ≻ t -> ren_tm ξ s ≻ @ren_tm n m ξ t.
Proof. 
  intros h. revert m ξ. induction h; intros m ξ; try (asimpl; constructor; eauto).
       - asimpl. apply substitutivity_h. asimpl. reflexivity.
Qed.


Lemma star_subst {n m} s t (σ : fin n -> tm m) : s ≻* t -> subst_tm σ s ≻* subst_tm σ t. 
Proof.
  intros H. revert σ. induction H; intros σ; eauto using srefl, strans, substitutivity.
Qed.

Lemma star_ren {n m} s t (ξ : fin n -> fin m) : s ≻* t -> ren_tm ξ s ≻* ren_tm ξ t.
Proof.
  intros H. revert ξ. induction H; intros ξ; eauto using srefl, strans, substitutivity_ren. 
Qed.

Lemma eval_appl {n} (e1 : tm n) e1' e2 : e1 ≻* e1' -> (app e1 e2) ≻* (app e1' e2).
Proof.
  induction 1. constructor. eauto using strans, eappl.
Qed.

Lemma eval_appr {n} (e1 : tm n) e2 e2' : e2 ≻* e2' -> (app e1 e2) ≻* (app e1 e2').
Proof.
  induction 1. constructor. eauto using strans, eappr.
Qed.

Lemma eval_app {n} (e1 : tm n) e1' e2 e2' : e1 ≻* e1' -> e2 ≻* e2' -> (app e1 e2) ≻* (app e1' e2'). 
Proof.
  intros. eauto using star_trans, eval_appr, eval_appl.
Qed.

Lemma eval_plus {n} (e1 : tm n) e1' e2 e2' : e1 ≻* e1' -> e2 ≻* e2' -> plus e1 e2 ≻* plus e1' e2'.
Proof.
  intros h1. revert e2 e2'. induction h1 as [e | e e1 e1' He1 He2 IH];
  intros e2 e2' h2.
  - induction h2; econstructor; eauto using strans, eplusr.
  - induction h2; econstructor; eauto using strans, eplusl, srefl.
Qed.

Lemma eval_lam {n} (e1 : tm (S n)) e2 : e1 ≻* e2 -> lam e1 ≻* lam e2.
Proof. induction 1. apply srefl. econstructor.  apply elam. exact H. eauto.
Qed.  


Lemma subst_eval {n m} s (σ : fin n -> tm m) τ : (forall x, σ x ≻* τ x) -> subst_tm σ s ≻* subst_tm τ s.
Proof. revert m σ τ. 
  induction s; intros m σ τ H; asimpl; eauto using eval_app, eval_plus, srefl.
       - asimpl. apply eval_lam. apply IHs. intros []; cbn.
         + unfold funcomp. specialize (H f). inversion H; eauto using star_ren, strans, srefl.
         + apply srefl.  
Qed.

Lemma subst_eval_head {n} (s : tm (S n)) t t' : t ≻ t' -> subst_tm (t.:var_tm) s ≻* subst_tm (t'.:var_tm) s.
Proof.
  intros Ht. apply subst_eval. intros []; cbn; eauto using srefl, strans.
Qed.

 
Lemma preservation {n} Γ (s : tm n) t : s ≻ t -> forall A, Γ ⊢ s : A -> Γ ⊢ t : A.
Proof.
  intros h. induction h; intros; inversion H; subst; try econstructor; eauto using H.
  + inversion H3 ; subst. 
    eapply morphism_subst. exact H4.
    intros []; cbv; eauto. constructor.
Qed.

Lemma agree_weaken {n m} (Γ : ctx n) (Δ : ctx m) A ξ : Γ ≼ Δ ; ξ -> Γ ≼ A.:Δ ; (ξ>> shift).
Proof.
  intros H x. rewrite <- (H x). now cbn.
Qed.

Lemma types_ren {n m} Γ (s : tm n) (ξ: fin n -> fin m) (Δ : ctx m) A : Γ ⊢ s : A -> Γ ≼ Δ ; ξ -> Δ ⊢ (ren_tm ξ s) : A.
Proof.
  intros H1.
  revert Δ ξ. revert m.
 induction H1; intros m' Δ ξ H2; asimpl; try (econstructor); eauto.
  - rewrite <- (H2 x). eauto using tvar.
  - apply IHtypes. intros [x|]; cbn; eauto. 
Qed.