Add LoadPath "~/ba/casestudy".
Require Export stlc_num.
Require Export wn.
Require Export eval.

Inductive SN {n} (R : tm n -> tm n -> Prop) : tm n -> Prop := 
| sn_step (e1 : tm n) : (forall e2, R e1 e2 -> SN R e2) -> SN R e1.
Print SN_ind. 
Definition SNe {n} := @SN n eval.

Lemma closed_SN_star {n} s t: s ≻* t -> SNe s -> @SNe n t .
Proof.
  intros H1 H2. induction H1; subst; eauto. apply IHstar.
  inversion H2; subst. now apply H0.
Qed.


Fixpoint R {n} (Γ : ctx n) A : tm n -> Prop :=
  match A with
  | int => fun s => (Γ ⊢ s : int) /\ SNe s
  | arrow A B => fun s => Γ ⊢ s : arrow A B /\
                              forall {m} ξ Δ t,
                                Γ ≼ Δ ; ξ -> R Δ A t -> @R m Δ B (app (ren_tm ξ s) t)
  end.

Definition agree_subst_sn {n m} (Γ : ctx n) (Δ : ctx m) (σ : fin n -> tm m) :=
  forall (x : fin n) , R Δ (Γ x) (σ x). 
Notation "Δ ≼ Γ , σ" := (agree_subst_sn Γ Δ σ) (at level 70).

Definition neutral {n} (s : tm n) : Prop :=
  match s with
  | lam s => False
  | _ => True
  end.

Lemma neutral_ren {n m} (ξ : fin n -> fin m) s : neutral s -> neutral (ren_tm ξ s). 
Proof.
  intros H. induction s; cbn in H; exact H. 
Qed.


Lemma types_R {n} (s : tm n) Γ A : R Γ A s -> Γ ⊢ s : A.
Proof.
  intros H. induction A; cbn in *; tauto.
Qed.

Lemma SN0 {n} x : @SNe n (var_tm x). 
Proof.
  econstructor. intros. inversion H; subst; contradiction H.
Qed.

Lemma types_agreement {n m} (Γ : ctx m) (Δ : ctx n) (σ : fin n -> tm m) :
  Γ ≼ Δ, σ -> forall (x : fin n) , Γ ⊢ (σ x) : (Δ x). 
Proof.
  intros H x. now apply types_R.
Qed.

Theorem morphism_subst_sn  {m} (Γ : ctx m) (s : tm m) A :
  Γ ⊢ s : A -> forall m' (σ: fin m -> tm m') Δ,
      (Δ ≼ Γ , σ) -> Δ ⊢ (subst_tm σ s) : A.
Proof.
  intros H1 m' σ Γ' H2. 
  pose (H := types_agreement _ _ _ H2).
  eapply morphism_subst; eauto.  
Qed.


Lemma SN1 {n} s t : SNe (app s t) -> SNe s /\ @SNe n t.
Proof.
  intros H. split.
  - remember (app s t) as st. revert s t Heqst.
    induction H as [st Hst IHst].
    intros s t Heqst. subst.
    constructor. intros s' Hs'. eapply (IHst (app s' t)); eauto using eappl.
  - remember (app s t) as st. revert s t Heqst.
    induction H as [st Hst IHst].
    intros s t Heqst. subst. constructor. intros t' Ht'. eapply (IHst (app s t')); eauto using eappr.
Qed.

Lemma SN1' {n} s t : SNe (app s t) -> @SNe n s.
  Proof. intros H. enough (SNe s /\ SNe t) by tauto. eauto using SN1. 
Qed.

Lemma SN2 {n m} s (σ : fin n -> tm m) : SNe (subst_tm σ s) -> @SNe n s.
Proof.
  intros H. remember (subst_tm σ s) as sσ.
  revert n s σ Heqsσ. induction H as [sσ Hsσ IHsσ].
  intros n s σ Heqsσ. subst. 
  constructor. intros s' Hs'.
  inversion Hs'; subst.
  (*; try (eapply IHsσ; econstructor; asimpl; eauto using substitutivity).*)
  - eapply IHsσ; constructor; asimpl; eauto using substitutivity.
  - eapply IHsσ; constructor; asimpl; eauto using substitutivity.
  - eapply IHsσ; constructor; asimpl; eauto using substitutivity.
  - eapply IHsσ; constructor; asimpl; eauto using substitutivity.
  - eapply IHsσ; constructor; asimpl; eauto using substitutivity.
  -  constructor; asimpl. intros c Hc.  inversion Hc. 
  - eapply (IHsσ (subst_tm σ (subst_tm (e2.: var_tm) e1))); eauto using substitutivity.  
 Qed.     
         
Lemma SN31 {n m} s (ξ : fin n -> fin m) : SNe (ren_tm ξ s) -> SNe s.
Proof.
  substify. apply SN2.
Qed.

Lemma const_inj {n} a b (H : @const n a = const b) : a = b.
Proof.
  change ((fun (x : tm n) => match x with | (const y) => y = b | _ => False end) (const a)).
  now rewrite  H.
Qed.

Lemma SN32' {n m} s t (ξ : fin n -> fin m) : ren_tm ξ s ≻ t -> exists t', t= ren_tm ξ t' /\ s ≻ t'.
Proof.
  revert m ξ t.
  induction s; intros m ξ t H.
  - asimpl in H; inversion H; subst.
  - asimpl in H; inversion H; subst.
    + destruct (IHs1 m ξ e1') as [s [Hs1 Hs2]]. auto.
      exists (app s s2). asimpl. rewrite Hs1.
      split; eauto using eappl.
    + destruct (IHs2 m ξ e2') as [s [Hs1 Hs2]]. auto.
      exists (app s1 s). asimpl. rewrite Hs1. 
      split; eauto using eappr.
    + destruct s1;
      try (asimpl in H; discriminate H1). asimpl in H. 
      exists (subst_tm (s2.:var_tm) s1). asimpl in H1.
      assert (H1' : e1 = ren_tm (bound.:ξ>> shift) s1).
      { change ((fun x => match x with| (lam s) => s=(ren_tm (bound.:ξ>> shift) s1) | _ => False end) (lam e1)).
      now rewrite H1. }  
      split.
      * rewrite H1'. now asimpl. 
      * econstructor. 
  - inversion H; subst.
    destruct (IHs (S m) (upRen_tm_tm ξ) e2) as [t [Ht1 Ht2]]; eauto.
    exists (lam t). asimpl in *. split. now rewrite Ht1.
    constructor. auto.
  - asimpl in *. inversion H. 
  - simpl in H. inversion H; subst.
    + destruct (IHs1 m ξ e1') as [s [Hs1 Hs2]]. eauto.
      exists (plus s s2). asimpl. rewrite Hs1.
      split; eauto using eplusl.
    + destruct (IHs2 m ξ e2') as [s [Hs1 Hs2]]. eauto.
      exists (plus s1 s). asimpl. rewrite Hs1. 
      split; eauto using eplusr.
    + destruct s1; destruct s2; try (discriminate H1); try (discriminate H2).
      exists (const (n+n0)).
      split. asimpl in *. pose (H3 := const_inj n1 n H1). pose (H4 := const_inj n2 n0 H2). 
      now rewrite H3, H4.
      econstructor. 
Qed.       
  
Lemma SN32 {n m} s (ξ : fin n -> fin m) : SNe s -> SNe (ren_tm ξ s) .
Proof.
  intros Hs. revert m ξ . induction Hs as [s Hs IHs]. 
  intros.
  constructor. intros s' Hs'.
  pose (H := SN32' s s' ξ Hs').
  destruct H as [t [Ht Ht']].
  rewrite Ht.
  apply IHs. auto.
Qed.

Lemma SN4 {n}  s t : SNe s -> SNe t -> SNe (subst_tm (t .: var_tm) s) -> @SNe n (app (lam s) t).
Proof.
  intros H. revert t. induction H as [s Hs IHs].
  intros t H. revert s Hs IHs.
  induction H as [t Ht IHt].
  intros. econstructor. intros s' Heval.
  inversion Heval; subst. inversion H3; subst.
  - apply IHs. eauto. constructor. apply Ht.
    inversion H as [t0 H0]. subst t0.
    apply H0. apply substitutivity. exact H1.
  - apply IHt; eauto. inversion H as [t0 H0]. subst t0.
    eapply closed_SN_star.
    apply subst_eval_head. exact H3. exact H.
  - auto. 
Qed.


Lemma SNplus {n} s t : SNe s -> SNe t -> @SNe n (plus s t).
Proof.
  intros H. revert t.
  induction H as [s Hs IHs].
  intros t H'. revert s Hs IHs.
  induction H' as [t Ht IHt]. intros. econstructor .
  intros s' H. inversion H; subst. eapply IHs.
  assumption.  econstructor. assumption. apply IHt. assumption. assumption. assumption.
  constructor. intros e He. inversion He; subst.
Qed.


Lemma SN_appzero {n} (s : tm n) :
  SNe (app (ren_tm shift s) (var_tm var_zero)) -> SNe s.
Proof.
  intros H. inversion H; subst. 
  eauto using SN31, SN1'. 
Qed.


Lemma R_ren {n m} (ξ : fin n -> fin m) Γ Δ A s :
  R Γ A s -> Γ ≼ Δ ; ξ -> R Δ A (ren_tm ξ s).
Proof. revert s Γ Δ ξ . revert n m. 
  induction A; intros n m s Γ Δ ξ Hs Hxi.
  - split. eauto using types_ren, types_R.
    apply SN32. destruct Hs as [_ Hs]. auto.
  - split. eauto using types_ren, types_R.
    intros m' ρ Δ' t Hrho Ht.
    asimpl. destruct Hs as [Hs1 Hs2].
    apply Hs2; eauto.
    intro x. unfold funcomp. rewrite <- (Hxi x). now rewrite (Hrho(ξ x)). 
Qed.

Lemma CR2' {n} Γ A (s : tm n) t : R Γ A s -> s ≻ t -> R Γ A t.
Proof.
  revert Γ s t. revert n. induction A.
  - intros n Γ s t [h1 h2] h3. split.
    + eapply preservation; eauto.
    + constructor. intros.
      inversion h2; subst. constructor. intros.
      apply (closed_SN_star t e0).
      * eauto using strans, srefl.
      * apply H0. eauto.
  - intros n Γ s t h1 h2. split.
    + eapply preservation; eauto. cbn in h1; tauto.
    + intros. destruct h1 as [r1 r2].
      specialize (r2 m ξ Δ t0 H H0).
      apply (IHA2 m Δ (app (ren_tm ξ s) t0)); eauto.
      econstructor. apply substitutivity_ren. assumption.
Qed. 
            
Lemma CR2 {n} Γ A (s : tm n) t : R Γ A s -> s ≻* t -> R Γ A t.
Proof.
  intros h1 h2. induction h2; eauto using CR2'. 
Qed.

Lemma CR13 {n} Γ A (s : tm n) :
  (R Γ A s -> SNe s) /\ (Γ ⊢ s : A -> neutral s -> (forall t, s ≻ t -> R Γ A t) -> R Γ A s).
Proof. 
  revert s Γ. revert n. induction A.
  - split.
    + intros [_ h2]. eauto.
    + intros h1 h2 h3.
      split; eauto. econstructor. intros e H.
      destruct (h3 e) as [h4 h5]; eauto.
  - split.
    + intros [h1 h2].
      apply SN_appzero.
      destruct (IHA2 (S n) (app (ren_tm shift s) (var_tm var_zero)) (A1.:Γ)) as [IHA21 IHA22].
      apply IHA21. apply h2.
      * intros x. now asimpl.
      * apply IHA1.
        -- econstructor.
        -- now cbn. 
        -- intros t H; inversion H.
    + intros h1 h2 h3. split; eauto.
      intros m ξ Δ t Hctx Hr.
      assert (Ht : SNe t).
      { destruct (IHA1 m t Δ); eauto. }
      induction Ht as [t Ht IHt]. apply IHA2.
      * inversion h1; subst; try (discriminate h1); try (contradiction h2).
        -- econstructor.
           ++ cbn. specialize (Hctx x).
           rewrite <- H2. rewrite <- Hctx. eapply tvar.
           ++ eauto using types_R.
        -- econstructor.
           ++ asimpl. econstructor; eauto using types_ren.
           ++ eauto using types_R.
      * now cbn.
      * intros t' Ht'. inversion Ht'; subst.
        -- destruct (SN32' s e1' ξ) as [s' [Hs1 Hs2]]; eauto. 
           rewrite Hs1. apply h3; eauto.  
        -- apply IHt. eauto. eapply CR2; eauto using srefl, strans.
        -- assert (Hn : ~(neutral (ren_tm ξ s))). rewrite <- H0. now cbn. 
               exfalso. apply Hn. eauto using neutral_ren. 
Qed.         

Lemma CR1 {n} Γ A (s : tm n) : R Γ A s -> SNe s.
Proof. apply CR13. Qed.


Lemma CR3 {n} Γ A (s : tm n) : Γ ⊢ s : A -> neutral s ->
                                      (forall t, s ≻ t -> R Γ A t) -> R Γ A s.
Proof. apply CR13. Qed.


Lemma R_varzero{n} (Γ : ctx n) A : R (A.:Γ) A (var_tm var_zero).
Proof.
  eapply CR3.
  - econstructor.
  - now cbn. 
  - intros t H; inversion H. 
Qed.

Lemma R_agreement {n m } (Γ : ctx m) (Γ' : ctx n) σ A : Γ ≼ Γ', σ -> (A.:Γ) ≼ (A.:Γ'), ((var_tm var_zero).: σ >> (ren_tm shift)). 
Proof.
  intros H [x|]; cbn.
  - specialize (H x). eapply R_ren. eauto.
    + now intros y. 
  - apply R_varzero.  
Qed.     
    
Lemma R_lam' {n} (s : tm (S n)) Γ A B t:
  R Γ A t -> SNe s -> (A.:Γ) ⊢ s : B -> R Γ B (subst_tm (t.: var_tm ) s ) -> R Γ B (app (lam s) t).
Proof.
  intros H1.
  assert (Ht : SNe t) by eauto using CR1.
  revert H1. revert A B s Γ.
  induction Ht as [t Ht IHt].
  intros A B s Γ H1 Hs. revert H1. revert Γ A B Ht IHt. revert t.
  induction Hs as [s Hs IHs].
  intros. 
  eapply CR3.
  - eauto using tapp, tlam, types_R.
  - now cbn.
  - intros t' Ht'. inversion Ht'; subst.
    + inversion H5; subst.
      eapply IHs; eauto using preservation.
     eapply CR2; eauto using substitutivity, strans, srefl.
    + eapply IHt; eauto using CR2, srefl, strans, star_trans.
      econstructor. exact Hs.
      eapply CR2. eauto. eauto using subst_eval_head.
    + eauto. 
Qed.      

Lemma soundness {n m} Γ Δ (σ : fin n -> tm m) s A :
  Γ ⊢ s : A -> ( Δ ≼ Γ , σ) ->  R Δ A (subst_tm σ s). 
Proof.
  intros H. revert m σ Δ.
  induction H; intros.
  - asimpl. apply (H x).  
  - asimpl. destruct (IHtypes1 _ σ Δ H1) as [H3 H4].
    specialize (H4 _ (id) Δ). asimpl in H4.
    apply H4. intros x. now asimpl.
    apply IHtypes2. assumption.
  -  split.  eapply morphism_subst_sn.
     econstructor; eauto. eauto. 
     destruct (IHtypes1 _ σ Δ H1 ) as [H3 H4].
     destruct (IHtypes2 _ σ Δ H1 ) as [H3' H4'].
     eapply CR1. apply CR3.
     asimpl. constructor; eauto. now cbn. 
     intros.
     inversion H2; subst.
     + split. * constructor; eauto.
     eauto using preservation.
     * apply SNplus; eauto. 
     destruct (IHtypes1 _ σ Δ H1) as [iht11 iht12].
     inversion iht12; subst. apply H5. eauto.
     + split; eauto using tplus, preservation.
     apply SNplus; eauto. destruct H4'. apply H5. eauto.
     + split; eauto using tint. econstructor; intros n Hn ;inversion Hn. 
  - asimpl. split.
    + econstructor. eapply morphism_subst_sn.
      * exact H.
      * eauto using R_agreement.
    + intros m' ξ Δ' t Hxi Ht. 
      asimpl. eapply R_lam'.
      * exact Ht.
      * eapply CR1; eauto. eapply IHtypes.
        intros [x|].
        -- cbn. unfold funcomp. eapply R_ren. 
           exact (H0 x). eauto using agree_weaken.
        -- cbn. apply R_varzero. 
          
      * apply types_R.
        eapply IHtypes.
        intros [x|].
        -- cbn. unfold funcomp. eapply R_ren. exact (H0 x).
           eauto using agree_weaken.
        -- cbn. apply R_varzero.
      * asimpl. 
        apply IHtypes. intros [x|]; cbn; eauto. 
        unfold funcomp. specialize (H0 x).
        unfold agree in Hxi.
        renamify.
        eauto using R_ren. 
  - split; asimpl; econstructor. intros e' H'. inversion H'; subst.
Qed.    
    
Lemma sn A s: ∅ ⊢ s : A -> SNe s.  
Proof.
  intros. eapply SN2.  apply (CR1 ∅ A). apply ( soundness ∅ ∅ var_tm ). eauto. intros [].
Qed.