import .sn_predicate
open tm
open type
open nat
open eval
open star

/- Logical redcibility relation R.-/
definition R : Π {n}, (ctx n) → type → tm n → Prop 
| n Γ tint s :=  (Γ ⊢ s : tint) ∧ SNe s
| n Γ (tarrow A B) s := Γ ⊢ s : tarrow A B ∧ forall {m : nat} ξ Δ t,
    (Γ ≼ Δ : ξ) → @R m Δ A t -> R Δ B (app (ren_tm ξ s) t)


definition agree_subst {n m} (Γ : ctx n) (Δ : ctx m) (σ : Fin n → tm m) := 
  ∀ (x : Fin n), R Δ (Γ x) (σ x).

notation Δ `≼`:50 Γ `:.`:50 σ :=  agree_subst Γ Δ σ. 

@[simp]
definition neutral {n} : tm n → Prop
| (lam s) := ff
| _ := tt 

lemma neutral_ren {n m} (ξ : Fin n → Fin m) (s : tm n) : neutral s → neutral (s.⟨ξ⟩) :=
begin 
  intro h, induction s; simp with substLemmas at*; aauto, 
end


lemma types_R {n} (s : tm n) (Γ A) : R Γ A s → Γ ⊢ s : A :=
begin 
  intro h, induction A; rw[R] at *; exact h.left,  
end

open types
meta def HintsTypes := [``(tvar), ``(tapp), ``(tplus), ``(tlam), ``(tint), 
``(types_R), ``(preservation), ``(types_ren), ``(agree_subst_types)]


lemma types_agreement {n m} (Γ : ctx m) (Δ : ctx n) (σ : Fin n → tm m) :
  (Γ ≼ Δ :. σ) → forall (x : Fin n), Γ ⊢ (σ x) : (Δ x) :=
begin 
  intros h x, apply types_R, aauto
end

--Morphism for agreement defined with R
theorem morphism_subst_sn {m} (Γ : ctx m) (s : tm m) (A) :
  Γ ⊢ s : A → forall m' (σ : Fin m → tm m') Δ , (Δ ≼ Γ :. σ) → Δ ⊢ s[σ] : A :=
begin 
  intros h₁ m' σ Δ h₂, 
  apply morphism_subst; aauto_w ``(types_agreement), 
end


lemma R_ren {n m} (ξ : Fin n → Fin m) (Γ) (Δ) (A) (s) : 
    R Γ A s → (Γ ≼ Δ : ξ) → R Δ A (s.⟨ ξ ⟩)  :=
begin 
    revert s Γ Δ ξ n m, induction A; intros n m s Γ Δ ξ hs hxi, 
    { split,
    try { do let l :=  [``(types_ren), ``(types_R)], aautow l , aautow l }, 
        apply SN₃l, exact hs.right },
    { split , apply types_ren, aauto_w ``(types_R), aauto,
    intros m' ρ Δ' t  hrho ht,  arw, 
    apply hs.right, intro x, unfold comp, rw ← hxi x, rw hrho, aauto }
end 


/- Reducibility properties in Girard's proof: CR₁, CR₂ and CR₃ -/
lemma CR₂' {n} (Γ) (A) (s t : tm n) : R Γ A s → s ≻ t → R Γ A t :=
begin   
    revert Γ s t n, induction A,
    { intros n Γ s t h q, split,
        { apply preservation; aauto_w ``(types_R), }, 
        constructor, intros t' ht', cases h.right, 
        apply closed_star, apply star.sstep, exact ht',
        aauto_w ``(star.srefl), apply a, aauto,
        }, 
    { intros n Γ s t h1 h2, split, 
        { apply preservation, aauto, apply h1.left,  },
        { intros n ξ Δ t p1 p2 , have p3 := (h1.right ξ Δ t p1 p2), 
        apply A_ih_a_1, assumption, constructor, aauto_w ``(substitutivity_ren),  }  }
end 

theorem CR₂ {n} (Γ) (A) (s t : tm n) : R Γ A s → s ≻* t → R Γ A t := 
begin 
  intros h q, induction q, 
  { aauto, }, 
  { apply q_ih, apply (CR₂'); aauto } 
end

theorem CR₁₃ {n} (Γ) (A) (s:tm n) : (R Γ A s → SNe s) ∧  
  (Γ ⊢ s : A → neutral s → (∀ t, s ≻ t → R Γ A t)  → R Γ A s) := 
begin
    revert s Γ n, induction A; intros,
  { split, 
    { intro h, apply h.right,}, 
    { intros, split; aauto, constructor, intros,
    apply (a_2 _ a_3).right
    },
  },
  { split, 
    { intro h, apply SN_appzero, 
    have p:= (A_ih_a_1 (scons A_a Γ) (app (ren_tm shift s) (var_tm var_zero))),  
    apply p.left, apply h.right, 
      { intro x,  refl, },
      { apply (A_ih_a _ _).right, 
        { constructor, },
        { simp, },
        { intros t q, cases q, },
      },
    },
    { intros h₁ h₂ h₃, split, assumption,
    intros m ξ Δ t p₁ p₂,
    have p₃ : SNe t, 
      from begin 
        apply (A_ih_a _ _).left, aauto,
      end,
    induction p₃, 
    apply (A_ih_a_1 _ _).right, 
      { generalize hb :(type.tarrow A_a A_a_1) = B,
        rw hb at h₁, 
        cases h₁, 
        any_goals { cases h₁; trivial,},
        { simp with substLemmas, constructor, 
          { rw hb at *, rw ← (p₁ (h₁_x)), constructor},
          { aauto_w ``(types_R) },
        },
      
        { constructor, 
          { simp with substLemmas, constructor, 
          apply types_ren, rw ← hb at * , assumption, assumption,
          aauto_w ``(types_ren) },
          { aauto_w ``(types_R) }, 
        },
      },
      
      { simp, },
      
      { intros t' ht', 
      generalize hs : s.⟨ξ⟩ = hxi, rw hs at ht', cases ht'; rw ← hs at *, 
        { have hsns := SN₃l' s ht'_e₁' ξ ht'_a, 
        cases hsns with e he, cases he with he₁ he₂,
        rw he₁, simp [R] at *, 
        apply ((h₃ _ _).right); aauto, },
        { aauto, apply CR₂, 
        repeat { aautow HintsStar }, },
        { have hnn : not (neutral (s.⟨ ξ ⟩)), 
        from begin rw hs, simp at *, end,
        exfalso, aauto, aauto_w ``(neutral_ren)
        }, 
      }, 
    },
  },
end

theorem CR₁ {n} (Γ) (A) (s : tm n) : R Γ A s → SNe s :=
begin
  apply (CR₁₃ _ _ _).left ,
end

theorem CR₃ {n} (Γ) (A) (s : tm n) : 
  Γ ⊢ s : A → neutral s → (∀ t, s ≻ t → R Γ A t) → R Γ A s :=
begin
  apply (CR₁₃ _ _ _).right,
end

lemma R_bound {n} (Γ : ctx n) (A) : R (scons A Γ) A (var_tm var_zero) :=
begin
  apply CR₃, 
  { constructor },
  { simp },
  { intros _ h, cases h },
end


lemma R_agreement {n m} (Γ:ctx  n) (Δ: ctx m) (σ : Fin n → tm m) (A : type) :
  (Δ ≼ Γ :. σ) → 
  ((scons A Δ) ≼ (scons A Γ) :.
  (scons(var_tm var_zero) (comp (ren_tm shift) σ)))  :=
  begin
    intros h x, cases x,
    { unfold scons,  apply (R_bound), 
   },
    { specialize (h x_a), simp with defs,
     apply R_ren, assumption, intro x, refl,  }, 

  end


lemma R_lam' {n} (s : tm (succ n)) (Γ A B t) :
  R Γ A t → SNe s → (scons A Γ) ⊢ s : B → R Γ B (s[t..]) 
    → R Γ B (app (lam s) t) :=
begin 
  intro h₁,
  have ht : SNe t, 
  from by aauto_w ``(CR₁),
  revert h₁ A B s Γ, 
  induction ht with t' ht iht,
  intros s Γ A B h₁ hs, revert h₁ A B Γ ht iht,
  induction hs with s' hs' ihs, intros,
  apply CR₃, 
    { apply tapp, apply tlam, assumption, apply types_R, assumption },
    { simp, },
    { intros t'' ht', 
    cases ht', 
      { cases ht'_a, 
      apply ihs; aauto_w ``(preservation),
      apply CR₂; aauto, 
      apply sstep, apply substitutivity; aauto, apply srefl, 
      },
      { apply iht, aauto, apply  CR₂; aauto, apply ev_to_star, aauto, 
      constructor, exact hs', aauto, apply CR₂ ; aauto, aauto_w ``(subst_eval_head) 
      },
      aauto 
    },
end

/- Soundness/fundamental theorem-/

theorem soundness {n m} (Γ Δ) (σ : Fin n → tm m) (s A) : 
Γ ⊢ s : A →  (Δ ≼ Γ :. σ ) → R Δ A (s[σ]) := 
begin
  intro h, revert m σ Δ, 
  induction h; intros,
  { arw, apply (a h_x), },
  { arw, cases (h_ih_a σ Δ a), specialize (right id Δ),
   arw_at (`right), 
   apply right,
   intro x, now_arw,
   apply h_ih_a_1, assumption },
  { split, 
    { apply morphism_subst_sn; aauto, constructor; assumption, },
    cases (h_ih_a σ Δ a) with h₃ h₄,
    cases (h_ih_a_1 σ Δ a) with h₃' h₄', 
    apply CR₁, 
      { apply CR₃; simp with substLemmas, 
        { constructor; assumption; simp with substLemmas, },
        { intros t h₅,
        generalize h₅' : plus (h_e₁[σ]) (h_e₂[σ]) = p,
        rw h₅' at h₅, cases h₅; rw ← h₅' at * ; try {cases h₅'},
          { split, 
            { constructor; aauto, 
            apply preservation; aauto }, 
            { apply SNplus; aauto, 
            cases (h_ih_a σ Δ  a) with ih₁₁ ih₁₂,
            cases ih₁₂ with h₆; aauto, 
            },
          },
          { split; aauto_w ``(tplus); aauto_w ``(preservation),
          apply SNplus; aauto, 
          cases h₄' with h₇ , aauto, 
          },
          { split; aauto_w ``(types.tint), 
          constructor, intros a ha, cases ha, 
          },
        },
      },
  },
    
  { simp with substLemmas, split,
    { constructor, apply morphism_subst_sn; aauto_w ``(R_agreement), },
    { intros m' ξ Δ' t hxi ht,
      simp with substLemmas, 
      apply R_lam'; aauto,
      { apply CR₁;
      arw,
      apply h_ih, 
      intro x, cases x,
        { simp[scons], apply R_bound, aauto }, 
        { simp[comp, scons], apply R_ren, 
        aauto_w ``(agree_weaken), aauto, },
      }, 

      { apply types_R, 
      arw,
      apply h_ih,
      intro x, cases x,
      { simp[scons], apply R_bound,},
      { simp[comp,scons], apply R_ren,
        aauto_w ``(agree_weaken), aauto, },
      },
      { arw, 
        aauto, intro x, cases x,
        { arw, aauto },
        { simp [scons, comp], 
        specialize (a x_a), 
        renamify,
          { apply R_ren; aauto,  },
          { intro x, by refl}, }  
      },
    },
  },

  { split; arw, all_goals { constructor }, 
      { intros _ h₁, cases h₁, },
  },
      
end

/- Strong normalization.-/

lemma strong_normalization (A : type) (s) : empty_ctx ⊢ s : A → SNe s :=
begin 
  intro h, apply SN₂, apply CR₁, 
  apply (soundness empty_ctx empty_ctx var_tm); aauto,
  intro x, cases x
end
