section
open tactic
open well_founded_tactics

/-- like `tactic.mk_local_pis` -/
meta def tactic.mk_local_lams : expr → tactic (list expr × expr)
| (expr.lam n bi d b) := do
  p ← mk_local' n bi d,
  (ps, r) ← tactic.mk_local_lams (expr.instantiate_var b p),
  return ((p :: ps), r)
| e := return ([], e)

private meta def strip_psum_ctors : expr → expr
| `(psum.inl %%e) := strip_psum_ctors e
| `(psum.inr %%e) := strip_psum_ctors e
| e := e

private meta def find_first_discriminated_pattern : expr → option nat
| `(psigma.mk %%pat %%pats) :=
  if pat.is_app then some 0
  else (+1) <$> find_first_discriminated_pattern pats
| pat := if pat.is_app then some 0
         else none

meta def wf_single_arg (arg_idx_per_mut_fn : list nat := []) : well_founded_tactics := {
  dec_tac := do {
    unfold_wf_rel,
    unfold_sizeof,
    `[dsimp [psigma.fst,psigma.snd] { fail_if_unchanged := ff }],
    cancel_nat_add_lt,
    trivial_nat_lt
  },
  rel_tac := λ _ eqns, do
    arg_idx_per_mut_fn ← if arg_idx_per_mut_fn = [] then do
      eqn::_ ← pure eqns,
      (pat_vars, eqn) ← tactic.mk_local_lams eqn,
      expr.macro _ [lhs,_] ← pure eqn,
      expr.app f_mut mut_pats ← pure lhs,
      let mut_pats := strip_psum_ctors mut_pats,
      idx ← find_first_discriminated_pattern mut_pats,
      pure $ eqns.map (λ_, idx)
    else pure arg_idx_per_mut_fn,
    split,
    `[apply measure_wf],
    arg_idx_per_mut_fn.mmap' $ λ idx, do
      try `[refine λ p, psum.rec _ _ p],
      done <|> do {
        (list.range idx).mmap' $ λ _, `[refine λ p, (_ : _ → ℕ) p.snd],
        `[exact sizeof]
      }
  }
end
