{-# LANGUAGE FlexibleContexts      #-}
{-# LANGUAGE MultiParamTypeClasses #-}
module Autosubst.Names where

import           Autosubst.GenM
import           Autosubst.Syntax
import           Autosubst.Types
import           Data.List        as L

-- Contains the names of the later functions / inductive datatypes etc.

-- TODO: Sort.

-- Separation symbol
sep :: String
sep = "_"

-- Variable constructors
var_ :: TId -> String
var_ x = "var" ++ sep ++ x

constructor_ :: TId -> String
constructor_ c = c

-- Name for toVarRen functions
toVarRen_ :: TId -> String
toVarRen_ x = "toVarRen" ++ sep ++ x

-- Name of types for renaming/substitutions
renTy_ :: TId -> String
renTy_ x = "renTy" ++ sep ++ x

substTy_ :: TId -> String
substTy_ x = "substTy" ++ sep ++ x

ren_ :: TId -> String
ren_ x = "ren" ++ sep ++ x

-- Name for Casts
castRen_ :: TId -> TId -> String
castRen_ x y = "castRen" ++ sep ++ x ++ sep ++ y

-- Up Functions
upRen_ :: TId -> TId -> String
upRen_ x y = "upRen" ++ sep ++ x ++ sep ++ y

-- Definition of Substitution
cast_ :: TId -> TId -> String
cast_ x y = "cast" ++ sep ++ x ++ sep ++ y

toVar_ :: TId -> String
toVar_ x = "toVar" ++ sep ++ x

up_ ::  TId -> TId -> String
up_ x y = "up" ++ sep ++ x ++ sep ++ y

subst_ :: TId -> String
subst_ x = "subst" ++ sep ++ x

-- Name for predicate that two substitutions are eequivalent
eqRen_ :: TId -> String
eqRen_ x = "eqRen" ++ sep ++ x

eqSubst_ :: TId -> String
eqSubst_ x = "eqSubst" ++ sep ++ x

-- Lifting from x to y on component z
upId_ :: TId -> TId -> String
upId_ x y = "upId" ++ sep ++ x ++ sep ++ y

idSubst_ :: TId -> String
idSubst_ x = "idSubst" ++ sep ++ x

idSubst'_ :: TId -> String
idSubst'_ x = "idSubst'" ++ sep ++ x

congr_ :: TId -> String
congr_ c = "congr" ++ sep ++ c

congr'_ :: TId -> [Position] -> String -- bds explicit
congr'_ c s = "congr"++ sep ++ c ++ " " ++ concatMap (const $ "_ ") s

up_ren_ren_ :: String
up_ren_ren_ = "up_ren_ren"

up_ren_subst_ :: TId -> TId -> String
up_ren_subst_ x y = "up_ren_subst" ++ sep ++ x ++ sep ++ y

up_subst_ren_ :: TId -> TId -> String
up_subst_ren_ x y = "up_subst_ren" ++ sep ++ x ++ sep ++ y

up_subst_subst_ :: TId -> TId -> String
up_subst_subst_ x y = "up_subst_subst" ++ sep ++ x ++ sep ++ y

rinstInst_ :: TId -> String
rinstInst_ x = "rinst_inst" ++ sep ++ x

rinstInst'_ :: TId -> String
rinstInst'_ x = "rinst_inst'" ++ sep ++ x

--derived lemmas
rinstInstL_ :: TId -> String
rinstInstL_ x = "rinstInst" ++ sep ++ x

rinstId_ :: TId -> String
rinstId_ x = "rinstId" ++ sep ++ x

instId_ :: TId -> String
instId_ x = "instId" ++ sep ++ x

compComp_ :: TId -> String
compComp_ x = "compComp" ++ sep ++ x

compComp'_ :: TId -> String
compComp'_ x = "compComp'" ++ sep ++ x

compSR_ :: TId -> String
compSR_ x = "compRen" ++ sep ++ x

compSR'_ :: TId -> String
compSR'_ x = "compRen'" ++ sep ++ x

compRS_ :: TId -> String
compRS_ x = "renComp" ++ sep ++ x

compRS'_ :: TId -> String
compRS'_ x = "renComp'" ++ sep ++ x

compRR_ :: TId -> String
compRR_ x = "renRen" ++ sep ++ x

compRR'_ :: TId -> String
compRR'_ x = "renRen'" ++ sep ++ x

renComp_ :: TId -> String
renComp_ x = "renComp" ++ sep ++ x

renComp'_ :: TId -> String
renComp'_ x = "renComp'" ++ sep ++ x

renRen_ :: TId -> String
renRen_ x = "renRen" ++ sep ++ x

renRen'_ :: TId -> String
renRen'_ x = "renRen'" ++ sep ++ x

varL_ :: TId -> String
varL_ x = "varL" ++ sep ++ x

varLRen_ :: TId -> String
varLRen_ x = "varLRen" ++ sep ++ x
---

up_rinstInst_ :: TId -> TId -> String
up_rinstInst_ x y = "rinstInst_up" ++ sep ++ x ++ sep ++ y

upExtRen_ :: TId -> TId -> String
upExtRen_ x y = "upExtRen" ++ sep ++ x ++ sep ++ y

upExt_ :: TId -> TId -> String
upExt_ x y = "upExt" ++ sep ++ x ++ sep ++ y

ext_ :: TId -> String
ext_ x = "ext" ++ sep ++ x

ext'_ :: TId -> String
ext'_ x = "ext'" ++ sep ++ x

extRen_ :: TId -> String
extRen_ x = "extRen" ++ sep ++ x

extRen'_ :: TId -> String
extRen'_ x = "extRen'" ++ sep ++ x

compRenRen_ :: TId -> String
compRenRen_ x = "compRenRen" ++ sep ++ x

compRenRen'_ :: TId -> String
compRenRen'_ x = "compRenRen'" ++ sep ++ x

compRenSubst_ :: TId -> String
compRenSubst_ x = "compRenSubst" ++ sep ++ x

compRenSubst'_ :: TId -> String
compRenSubst'_ x = "compRenSubst'" ++ sep ++ x

compSubstRen_ :: TId -> String
compSubstRen_ x = "compSubstRen" ++ sep ++ x

compSubstRen'_ :: TId -> String
compSubstRen'_ x = "compSubstRen'" ++ sep ++ x

compSubstSubst_ :: TId -> String
compSubstSubst_ x = "compSubstSubst" ++ sep ++ x

compSubstSubst'_ :: TId -> String
compSubstSubst'_ x = "compSubstSubst'" ++ sep ++ x

comp'_ :: TId -> String
comp'_ x = "comp'" ++ sep ++ x

lidComp_ :: TId -> String
lidComp_ x = "lidComp" ++ sep ++ x

is_ :: String -> TId -> String
is_ s "" = "is" ++ sep ++ s
is_ s x = "is" ++ sep ++ s ++ sep ++ x

destruct_ :: String -> TId -> String
destruct_ s "" = "destruct" ++ sep ++ s
destruct_ s x = "destruct" ++ sep ++ s ++ sep ++ x

nmCmp_ :: TId -> String
nmCmp_ x = "normalize_"++x++" Comp"
nmCons_ :: TId -> String
nmCons_ x = "normalize_"++x++" Cons"
nmUp_ :: TId -> String
nmUp_ x = "normalize_"++x++" Up"
nmInst_ :: TId -> String
nmInst_ x = "normalize_"++x++" Inst"
nmRinst_ :: TId -> String
nmRinst_ x = "normalize_"++x++" Rinst"
