import .fintype
import .wftactic
run_cmd mk_simp_attr `substLemmas

@[substLemmas]
inductive ty  : nat -> Type 
  | var_ty : Π (nty : nat), Fin nty -> ty nty
  | arr : Π (nty : nat), ty nty -> ty nty -> ty nty
  | all : Π (nty : nat), ty (nat.succ nty) -> ty nty
open ty 

@[substLemmas]
lemma congr_arr { mty : nat } { s0 : ty mty } { s1 : ty mty } { t0 : ty mty } { t1 : ty mty } : s0 = t0 -> s1 = t1 -> arr mty s0 s1 = arr mty t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_all { mty : nat } { s0 : ty (nat.succ mty) } { t0 : ty (nat.succ mty) } : s0 = t0 -> all mty s0 = all mty t0 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_arr' { mty : nat } { s0 : ty mty } { s1 : ty mty } { t0 : ty mty } { t1 : ty mty } (h0 : s0 = t0) (h1 : s1 = t1) : arr mty s0 s1 = arr mty t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_all' { mty : nat } { s0 : ty (nat.succ mty) } { t0 : ty (nat.succ mty) } (h0 : s0 = t0) : all mty s0 = all mty t0 :=
begin intros, congr; assumption end

@[substLemmas]
def upRen_ty_ty { m } { n } (xi : Fin m -> Fin n) : _ :=
  up_ren xi

@[substLemmas]
def ren_ty : Π { mty : nat } { nty : nat } (xity : Fin mty -> Fin nty) (s : ty mty), ty nty
    | mty nty xity  (var_ty _ s) := (var_ty nty) (xity s)
    | mty nty xity  (arr _ s0 s1) := arr nty (ren_ty xity s0) (ren_ty xity s1)
    | mty nty xity  (all _ s0) := all nty (ren_ty (upRen_ty_ty xity) s0)

@[substLemmas]
def up_ty_ty { m } { nty : nat } (sigma : Fin m -> ty nty) : _ :=
  scons ((var_ty (nat.succ nty)) var_zero) (comp (ren_ty shift) sigma)

@[substLemmas]
def subst_ty : Π { mty : nat } { nty : nat } (sigmaty : Fin mty -> ty nty) (s : ty mty), ty nty
    | mty nty sigmaty  (var_ty _ s) := sigmaty s
    | mty nty sigmaty  (arr _ s0 s1) := arr nty (subst_ty sigmaty s0) (subst_ty sigmaty s1)
    | mty nty sigmaty  (all _ s0) := all nty (subst_ty (up_ty_ty sigmaty) s0)

@[substLemmas]
def upId_ty_ty { mty : nat } { sigma : Fin mty -> ty mty } (Eq : forall x, sigma x = (var_ty mty) x) : forall x, (up_ty_ty sigma) x = (var_ty (nat.succ mty)) x :=
  λ n , match n with
  | (some n) := have h : _, from ap (ren_ty shift) (Eq n), begin simp with substLemmas at *, assumption end
  | none := by refl
  end

@[substLemmas]
def idSubst_ty : Π { mty : nat } (sigmaty : Fin mty -> ty mty) (Eqty : forall x, sigmaty x = (var_ty mty) x) (s : ty mty), subst_ty sigmaty s = s
    | mty sigmaty Eqty  (var_ty _ s) := have h : _, from Eqty s, begin simp with substLemmas at *, assumption end
    | mty sigmaty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (idSubst_ty sigmaty Eqty s0) (idSubst_ty sigmaty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty sigmaty Eqty 
    (all _ s0) := have h : _, from congr_all (idSubst_ty (up_ty_ty sigmaty) (upId_ty_ty (_) Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma idSubst'_ty { mty : nat } { sigmaty : Fin mty -> ty mty } { Eqty : forall x, sigmaty x = (var_ty mty) x } { s s' : ty mty } (Eqty : forall x, sigmaty x = (var_ty mty) x) (Eqsty : s = s') : subst_ty sigmaty s = s' :=
eq.trans (idSubst_ty sigmaty Eqty s) Eqsty

@[substLemmas]
def upExtRen_ty_ty { m } { n } (xi : Fin m -> Fin n) (zeta : Fin m -> Fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_ty_ty xi) x = (upRen_ty_ty zeta) x :=
  λ n , match n with
  | (some n) := ap shift (Eq n)
  | none := by refl
  end

@[substLemmas]
def extRen_ty : Π { mty : nat } { nty : nat } (xity : Fin mty -> Fin nty) (zetaty : Fin mty -> Fin nty) (Eqty : forall x, xity x = zetaty x) (s : ty mty), ren_ty xity s = ren_ty zetaty s
    | mty nty xity zetaty Eqty  (var_ty _ s) := have h : _, from ap (var_ty nty) (Eqty s), begin simp with substLemmas at *, assumption end
    | mty nty xity zetaty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (extRen_ty xity zetaty Eqty s0) (extRen_ty xity zetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty nty xity zetaty Eqty 
    (all _ s0) := have h : _, from congr_all (extRen_ty (upRen_ty_ty xity) (upRen_ty_ty zetaty) (upExtRen_ty_ty xity zetaty Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma extRen'_ty { mty : nat } { nty : nat } { xity : Fin mty -> Fin nty } { zetaty : Fin mty -> Fin nty } { s s' : ty mty } (Eqty : forall x, xity x = zetaty x) (Eqsty : s = s') : ren_ty xity s = ren_ty zetaty s' :=
have h : _, from extRen_ty xity zetaty Eqty s, begin rw <- Eqsty, assumption end

@[substLemmas]
def upExt_ty_ty { m } { nty : nat } (sigma : Fin m -> ty nty) (tau : Fin m -> ty nty) (Eq : forall x, sigma x = tau x) : forall x, (up_ty_ty sigma) x = (up_ty_ty tau) x :=
  λ n , match n with
  | (some n) := ap (ren_ty shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
def ext_ty : Π { mty : nat } { nty : nat } (sigmaty : Fin mty -> ty nty) (tauty : Fin mty -> ty nty) (Eqty : forall x, sigmaty x = tauty x) (s : ty mty), subst_ty sigmaty s = subst_ty tauty s
    | mty nty sigmaty tauty Eqty  (var_ty _ s) := have h : _, from Eqty s, begin simp with substLemmas at *, assumption end
    | mty nty sigmaty tauty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (ext_ty sigmaty tauty Eqty s0) (ext_ty sigmaty tauty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty nty sigmaty tauty Eqty 
    (all _ s0) := have h : _, from congr_all (ext_ty (up_ty_ty sigmaty) (up_ty_ty tauty) (upExt_ty_ty sigmaty tauty Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma ext'_ty { mty : nat } { nty : nat } { sigmaty : Fin mty -> ty nty } { tauty : Fin mty -> ty nty } { s s' : ty mty } (Eqty : forall x, sigmaty x = tauty x) (Eqsty : s = s') : subst_ty sigmaty s = subst_ty tauty s' :=
have h : _, from ext_ty sigmaty tauty Eqty s, begin rw <- Eqsty, assumption end

@[substLemmas]
def compRenRen_ty : Π { kty : nat } { lty : nat } { mty : nat } (xity : Fin mty -> Fin kty) (zetaty : Fin kty -> Fin lty) (rhoty : Fin mty -> Fin lty) (Eqty : forall x, (comp zetaty xity) x = rhoty x) (s : ty mty), ren_ty zetaty (ren_ty xity s) = ren_ty rhoty s
    | kty lty mty xity zetaty rhoty Eqty 
    (var_ty _ s) := have h : _, from ap (var_ty lty) (Eqty s), begin simp with substLemmas at *, assumption end
    | kty lty mty xity zetaty rhoty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (compRenRen_ty xity zetaty rhoty Eqty s0) (compRenRen_ty xity zetaty rhoty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty lty mty xity zetaty rhoty Eqty 
    (all _ s0) := have h : _, from congr_all (compRenRen_ty (upRen_ty_ty xity) (upRen_ty_ty zetaty) (upRen_ty_ty rhoty) (up_ren_ren xity zetaty rhoty Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compRenRen'_ty { kty : nat } { lty : nat } { mty : nat } { xity : Fin mty -> Fin kty } { zetaty : Fin kty -> Fin lty } { rhoty : Fin mty -> Fin lty } { s : ty mty } { s' : ty kty } (Eqty : forall x, (comp zetaty xity) x = rhoty x) (Eqsty : s' = ren_ty xity s) : ren_ty zetaty s' = ren_ty rhoty s :=
have h : _, from compRenRen_ty xity zetaty rhoty Eqty s, begin rw Eqsty, assumption end

@[substLemmas]
def up_ren_subst_ty_ty { k } { l } { mty : nat } (xi : Fin k -> Fin l) (tau : Fin l -> ty mty) (theta : Fin k -> ty mty) (Eq : forall x, (comp tau xi) x = theta x) : forall x, (comp (up_ty_ty tau) (upRen_ty_ty xi)) x = (up_ty_ty theta) x :=
  λ n , match n with
  | (some n) := ap (ren_ty shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
def compRenSubst_ty : Π { kty : nat } { lty : nat } { mty : nat } (xity : Fin mty -> Fin kty) (tauty : Fin kty -> ty lty) (thetaty : Fin mty -> ty lty) (Eqty : forall x, (comp tauty xity) x = thetaty x) (s : ty mty), subst_ty tauty (ren_ty xity s) = subst_ty thetaty s
    | kty lty mty xity tauty thetaty Eqty  (var_ty _ s) := have h : _, from Eqty s, begin simp with substLemmas at *, assumption end
    | kty lty mty xity tauty thetaty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (compRenSubst_ty xity tauty thetaty Eqty s0) (compRenSubst_ty xity tauty thetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty lty mty xity tauty thetaty Eqty 
    (all _ s0) := have h : _, from congr_all (compRenSubst_ty (upRen_ty_ty xity) (up_ty_ty tauty) (up_ty_ty thetaty) (up_ren_subst_ty_ty xity tauty thetaty Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compRenSubst'_ty { kty : nat } { lty : nat } { mty : nat } { xity : Fin mty -> Fin kty } { tauty : Fin kty -> ty lty } { thetaty : Fin mty -> ty lty } { s : ty mty } { s' : ty kty } (Eqty : forall x, (comp tauty xity) x = thetaty x) (Eqsty : s' = ren_ty xity s) : subst_ty tauty s' = subst_ty thetaty s :=
have h : _, from compRenSubst_ty xity tauty thetaty Eqty s, begin rw Eqsty, assumption end

@[substLemmas]
def up_subst_ren_ty_ty { k } { lty : nat } { mty : nat } (sigma : Fin k -> ty lty) (zetaty : Fin lty -> Fin mty) (theta : Fin k -> ty mty) (Eq : forall x, (comp (ren_ty zetaty) sigma) x = theta x) : forall x, (comp (ren_ty (upRen_ty_ty zetaty)) (up_ty_ty sigma)) x = (up_ty_ty theta) x :=
  λ n , match n with
  | (some n) := eq.trans (compRenRen_ty shift (upRen_ty_ty zetaty) (comp shift zetaty) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compRenRen_ty zetaty shift (comp shift zetaty) (λ x , by refl) (sigma n))) (ap (ren_ty shift) (Eq n)))
  | none := by refl
  end

@[substLemmas]
def compSubstRen_ty : Π { kty : nat } { lty : nat } { mty : nat } (sigmaty : Fin mty -> ty kty) (zetaty : Fin kty -> Fin lty) (thetaty : Fin mty -> ty lty) (Eqty : forall x, (comp (ren_ty zetaty) sigmaty) x = thetaty x) (s : ty mty), ren_ty zetaty (subst_ty sigmaty s) = subst_ty thetaty s
    | kty lty mty sigmaty zetaty thetaty Eqty  (var_ty _ s) := have h : _, from Eqty s, begin simp with substLemmas at *, assumption end
    | kty lty mty sigmaty zetaty thetaty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (compSubstRen_ty sigmaty zetaty thetaty Eqty s0) (compSubstRen_ty sigmaty zetaty thetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty lty mty sigmaty zetaty thetaty Eqty 
    (all _ s0) := have h : _, from congr_all (compSubstRen_ty (up_ty_ty sigmaty) (upRen_ty_ty zetaty) (up_ty_ty thetaty) (up_subst_ren_ty_ty sigmaty zetaty thetaty Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compSubstRen'_ty { kty : nat } { lty : nat } { mty : nat } { sigmaty : Fin mty -> ty kty } { zetaty : Fin kty -> Fin lty } { thetaty : Fin mty -> ty lty } { s : ty mty } { s' : ty kty } (Eqty : forall x, (comp (ren_ty zetaty) sigmaty) x = thetaty x) (Eqsty : s' = subst_ty sigmaty s) : ren_ty zetaty s' = subst_ty thetaty s :=
have h : _, from compSubstRen_ty sigmaty zetaty thetaty Eqty s, begin rw Eqsty, assumption end

@[substLemmas]
def up_subst_subst_ty_ty { k } { lty : nat } { mty : nat } (sigma : Fin k -> ty lty) (tauty : Fin lty -> ty mty) (theta : Fin k -> ty mty) (Eq : forall x, (comp (subst_ty tauty) sigma) x = theta x) : forall x, (comp (subst_ty (up_ty_ty tauty)) (up_ty_ty sigma)) x = (up_ty_ty theta) x :=
  λ n , match n with
  | (some n) := eq.trans (compRenSubst_ty shift (up_ty_ty tauty) (comp (up_ty_ty tauty) shift) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compSubstRen_ty tauty shift (comp (ren_ty shift) tauty) (λ x , by refl) (sigma n))) (ap (ren_ty shift) (Eq n)))
  | none := by refl
  end

@[substLemmas]
def compSubstSubst_ty : Π { kty : nat } { lty : nat } { mty : nat } (sigmaty : Fin mty -> ty kty) (tauty : Fin kty -> ty lty) (thetaty : Fin mty -> ty lty) (Eqty : forall x, (comp (subst_ty tauty) sigmaty) x = thetaty x) (s : ty mty), subst_ty tauty (subst_ty sigmaty s) = subst_ty thetaty s
    | kty lty mty sigmaty tauty thetaty Eqty  (var_ty _ s) := have h : _, from Eqty s, begin simp with substLemmas at *, assumption end
    | kty lty mty sigmaty tauty thetaty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (compSubstSubst_ty sigmaty tauty thetaty Eqty s0) (compSubstSubst_ty sigmaty tauty thetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty lty mty sigmaty tauty thetaty Eqty 
    (all _ s0) := have h : _, from congr_all (compSubstSubst_ty (up_ty_ty sigmaty) (up_ty_ty tauty) (up_ty_ty thetaty) (up_subst_subst_ty_ty sigmaty tauty thetaty Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compSubstSubst'_ty { kty : nat } { lty : nat } { mty : nat } { sigmaty : Fin mty -> ty kty } { tauty : Fin kty -> ty lty } { thetaty : Fin mty -> ty lty } { s : ty mty } { s' : ty kty } (Eqty : forall x, (comp (subst_ty tauty) sigmaty) x = thetaty x) (Eqsty : s' = subst_ty sigmaty s) : subst_ty tauty s' = subst_ty thetaty s :=
have h : _, from compSubstSubst_ty sigmaty tauty thetaty Eqty s, begin rw Eqsty, assumption end

@[substLemmas]
lemma comp'_ty { mty : nat } { nty : nat } { sigma'ty : Fin mty -> ty nty } { tau'ty : Fin mty -> ty nty } (Eqsty : forall x, sigmaty x = sigma'ty x) (Eqtty : forall x, sigmaty x = sigma'ty x) : forall x, (comp (subst_ty tauty) sigmaty) x = (comp (subst_ty tau'ty) sigmaty) x :=
begin assumption end

@[substLemmas]
def rinstInst_up_ty_ty { m } { nty : nat } (xi : Fin m -> Fin nty) (sigma : Fin m -> ty nty) (Eq : forall x, (comp (var_ty nty) xi) x = sigma x) : forall x, (comp (var_ty (nat.succ nty)) (upRen_ty_ty xi)) x = (up_ty_ty sigma) x :=
  λ n , match n with
  | (some n) := ap (ren_ty shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
def rinst_inst_ty : Π { mty : nat } { nty : nat } (xity : Fin mty -> Fin nty) (sigmaty : Fin mty -> ty nty) (Eqty : forall x, (comp (var_ty nty) xity) x = sigmaty x) (s : ty mty), ren_ty xity s = subst_ty sigmaty s
    | mty nty xity sigmaty Eqty  (var_ty _ s) := have h : _, from Eqty s, begin simp with substLemmas at *, assumption end
    | mty nty xity sigmaty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (rinst_inst_ty xity sigmaty Eqty s0) (rinst_inst_ty xity sigmaty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty nty xity sigmaty Eqty 
    (all _ s0) := have h : _, from congr_all (rinst_inst_ty (upRen_ty_ty xity) (up_ty_ty sigmaty) (rinstInst_up_ty_ty xity sigmaty Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma rinst_inst'_ty { mty : nat } { nty : nat } { sigmaty : Fin mty -> ty nty } { xity : Fin mty -> Fin nty } { s s' : ty mty } (Eqty : forall x, (comp (var_ty nty) xity) x = sigmaty x) (Eqsty : s = s') : ren_ty xity s' = subst_ty sigmaty s :=
have h : _, from rinst_inst_ty xity sigmaty Eqty s, begin rw <- Eqsty, assumption end

@[substLemmas]
lemma instId_ty { mty : nat } : subst_ty (var_ty mty) = id :=
 (λ x , idSubst_ty (var_ty mty) (λ n , by refl) (id x))

@[substLemmas]
lemma varL_ty { mty : nat } { nty : nat } (sigmaty : Fin mty -> ty nty) : comp (subst_ty sigmaty) (var_ty mty) = sigmaty :=
 (λ x , by refl)

@[substLemmas]
lemma rinstInst_ty { mty : nat } { nty : nat } (xity : Fin mty -> Fin nty) : ren_ty xity = subst_ty (comp (var_ty nty) xity) :=
 (λ x , rinst_inst_ty xity (_) (λ n , by refl) x)

@[substLemmas]
lemma compComp_ty { kty : nat } { lty : nat } { mty : nat } (sigmaty : Fin mty -> ty kty) (tauty : Fin kty -> ty lty) (s : ty mty) : subst_ty tauty (subst_ty sigmaty s) = subst_ty (comp (subst_ty tauty) sigmaty) s :=
compSubstSubst_ty sigmaty tauty (_) (λ n , by refl) s

@[substLemmas]
lemma compComp'_ty { kty : nat } { lty : nat } { mty : nat } (sigmaty : Fin mty -> ty kty) (tauty : Fin kty -> ty lty) : comp (subst_ty tauty) (subst_ty sigmaty) = subst_ty (comp (subst_ty tauty) sigmaty) :=
 (λ n , compComp_ty sigmaty tauty n)

meta def is_arr_ty : expr -> bool
  | `(arr _ %%s0 %%s1) := tt
  | `(congr_arr' %%s0 %%s1) := tt
  | `(idSubst'_ty %%Eqty %%Eqsty) := is_arr_ty Eqsty
  | _ := ff

meta def destruct_arr_ty : expr -> tactic (expr × expr)
  | `(arr _ %%s0 %%s1) := return (s0, s1)
  | `(congr_arr' %%s0 %%s1) := return (s0, s1)
  | `(idSubst'_ty %%Eqty %%Eqsty) := destruct_arr_ty Eqsty
  | e := return (e, e)

meta def is_all_ty : expr -> bool
  | `(all _ %%s0) := tt
  | `(congr_all' %%s0) := tt
  | `(idSubst'_ty %%Eqty %%Eqsty) := is_all_ty Eqsty
  | _ := ff

meta def destruct_all_ty : expr -> tactic expr
  | `(all _ %%s0) := return s0
  | `(congr_all' %%s0) := return s0
  | `(idSubst'_ty %%Eqty %%Eqsty) := destruct_all_ty Eqsty
  | e := return e

meta def is_ren_ty : expr -> bool
  | `(extRen'_ty %%Eqty %%Eqsty) := tt
  | _ := ff

meta def destruct_ren_ty : expr -> tactic (expr × expr)
  | `(extRen'_ty %%Eqty %%Eqsty) := return (Eqty, Eqsty)
  | `(%%e) := return (e,e)

meta def is_subst_ty : expr -> bool
  | `(ext'_ty %%Eqty %%Eqsty) := tt
  | _ := ff

meta def destruct_subst_ty : expr -> tactic (expr × expr)
  | `(ext'_ty %%Eqty %%Eqsty) := return (Eqty, Eqsty)
  | `(%%e) := return (e,e)

meta def is_InstRen_ty : expr -> bool
  | `(compRenRen'_ty %%Eqty %%Eqsty) := is_InstRen_ty Eqsty
  | _ := ff

meta def destruct_InstRen_ty : expr -> tactic (expr × expr)
  | `(compRenRen'_ty %%Eqty %%Eqsty) := destruct_InstRen_ty Eqsty
  | `(%%e) := return (e,e)

meta def is_Inst_ty : expr -> bool
  | `(compRenRen'_ty %%Eqty %%Eqsty) := is_Inst_ty Eqsty
  | `(compRenSubst'_ty %%Eqty %%Eqsty) := is_Inst_ty Eqsty
  | `(compSubstRen'_ty %%Eqty %%Eqsty) := is_Inst_ty Eqsty
  | `(compSubstSubst'_ty %%Eqty %%Eqsty) := is_Inst_ty Eqsty
  | `(idSubst'_ty %%Eqty %%Eqsty) := is_Inst_ty Eqsty
  | `(rinst_inst'_ty %%Eqty %%Eqsty) := is_Inst_ty Eqsty
  | _ := ff

meta def destruct_Inst_ty : expr -> tactic (expr × expr)
  | `(ext'_ty %%Eqty %%Eqsty) := return (Eqty, Eqsty)
  | `(compRenRen'_ty %%Eqty %%Eqsty) := destruct_Inst_ty Eqsty
  | `(compRenSubst'_ty %%Eqty %%Eqsty) := destruct_Inst_ty Eqsty
  | `(compSubstRen'_ty %%Eqty %%Eqsty) := destruct_Inst_ty Eqsty
  | `(compSubstSubst'_ty %%Eqty %%Eqsty) := destruct_Inst_ty Eqsty
  | `(idSubst'_ty %%Eqty %%Eqsty) := destruct_Inst_ty Eqsty
  | `(rinst_inst'_ty %%Eqty %%Eqsty) := destruct_Inst_ty Eqsty
  | `(%%e) := return (e,e)

meta def is_scons : expr -> bool
  | `(up_ren_subst_ty_ty %%xi %%tau %%theta %%Eq) := is_scons Eq
  | `(up_subst_ren_ty_ty %%sigma %%zetaty %%theta %%Eq) := is_scons Eq
  | `(up_subst_subst_ty_ty %%sigma %%tauty %%theta %%Eq) := is_scons Eq
  | `(scons %%H1 %%H2) := tt
  | _ := ff

meta def is_comp : expr -> bool
  | `(up_ren_subst_ty_ty %%xi %%tau %%theta %%Eq) := is_comp Eq
  | `(up_subst_ren_ty_ty %%sigma %%zetaty %%theta %%Eq) := is_comp Eq
  | `(up_subst_subst_ty_ty %%sigma %%tauty %%theta %%Eq) := is_comp Eq
  | `(comp %%H1 %%H2) := tt
  | _ := ff

meta def is_var_ty : expr -> bool
  | `(up_ren_subst_ty_ty %%xi %%tau %%theta %%Eq) := is_var_ty Eq
  | `(up_subst_ren_ty_ty %%sigma %%zetaty %%theta %%Eq) := is_var_ty Eq
  | `(up_subst_subst_ty_ty %%sigma %%tauty %%theta %%Eq) := is_var_ty Eq
  | `(var_ty %%H1 %%H2) := tt
  | _ := ff

meta def destruct_scons : expr -> tactic (expr × expr)
  | `(up_ren_subst_ty_ty %%xi %%tau %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_ren_ty_ty %%sigma %%zetaty %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_subst_ty_ty %%sigma %%tauty %%theta %%Eq) := destruct_scons Eq
  | `(scons %%H1 %%H2) := return (H1, H2)
  | `(%%e) := return (e,e)

meta def destruct_comp : expr -> tactic (expr × expr)
  | `(up_ren_subst_ty_ty %%xi %%tau %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_ren_ty_ty %%sigma %%zetaty %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_subst_ty_ty %%sigma %%tauty %%theta %%Eq) := destruct_comp Eq
  | `(comp %%H1 %%H2) := return (H1, H2)
  | `(%%e) := return (e,e)

meta def destruct_var_ty : expr -> tactic expr
  | `(up_ren_subst_ty_ty %%xi %%tau %%theta %%Eq) := destruct_var_ty Eq
  | `(up_subst_ren_ty_ty %%sigma %%zetaty %%theta %%Eq) := destruct_var_ty Eq
  | `(up_subst_subst_ty_ty %%sigma %%tauty %%theta %%Eq) := destruct_var_ty Eq
  | `(var_ty %%s) := return s
  | `(%%e) := return e

meta def normalize'_ty : nat -> expr -> expr -> tactic expr
  | n e1 e2 := if (n = Inst)
  then if (is_var_ty e2)
  then ex ``(idSubst'_ty %%e1 %%e2)
  else if (is_lam_ty e1)
  then do e3 ← destruct_lam_ty e1, e4 ← normalize'_ty Up e2 e2, e5 ← normalize'_ty Inst e3 e1, ex ``(congr_lam' %%e4 %%e5)
  else if (is_subst_ty e2)
  then do (e3,e4) ← destruct_subst_ty e2, e5 ← normalize'_ty Comp e4 e2, e6 ← normalize'_ty Inst e3 e5, ex ``(compSubstSubst'_ty %%e5 %%e1 %%e6)
  else ex ``(ext'_ty %%e1 %%e2)
  else if (n = Rinst)
  then if (is_var_ty e2)
  then ex ``(idSubst'_ty %%e1 %%e2)
  else if (is_lam_ty e1)
  then do e3 ← destruct_lam_ty e1, e4 ← normalize'_ty Up e2 e2, e5 ← normalize'_ty Rinst e3 e1, ex ``(congr_lam' %%e4 %%e5)
  else if (is_ren_ty e2)
  then do (e3,e4) ← destruct_ren_ty e2, e5 ← normalize'_ty Comp e4 e2, e6 ← normalize'_ty Rinst e3 e5, ex ``(compRenRen'_ty %%e5 %%e1 %%e6)
  else ex ``(extRen'_ty %%e1 %%e2)
  else if (n = Comp)
  then if (is_scons e1)
  then do (e3,e4) ← destruct_scons e1, e5 ← normalize'_ty Inst e3 e2, e6 ← normalize'_ty Comp e4 e2, e7 ← normalize'_ty Cons e5 e6, ex ``(scons_comp' %%e5 %%e1 %%e6)
  else ex ``(comp'_ty %%e1 %%e2)
  else if (n = Up)
  then TODO
  else ex ``(%%e1=%%e2)

@[substLemmas]
inductive tm  : nat -> nat -> Type 
  | var_tm : Π (nty ntm : nat), Fin ntm -> tm nty ntm
  | app : Π (nty ntm : nat), tm nty ntm -> tm nty ntm -> tm nty ntm
  | lam : Π (nty ntm : nat), ty nty -> tm nty (nat.succ ntm) -> tm nty ntm
  | tapp : Π (nty ntm : nat), tm nty ntm -> ty nty -> tm nty ntm
  | tlam : Π (nty ntm : nat), tm (nat.succ nty) ntm -> tm nty ntm
open tm 

@[substLemmas]
lemma congr_app { mty mtm : nat } { s0 : tm mty mtm } { s1 : tm mty mtm } { t0 : tm mty mtm } { t1 : tm mty mtm } : s0 = t0 -> s1 = t1 -> app mty mtm s0 s1 = app mty mtm t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_lam { mty mtm : nat } { s0 : ty mty } { s1 : tm mty (nat.succ mtm) } { t0 : ty mty } { t1 : tm mty (nat.succ mtm) } : s0 = t0 -> s1 = t1 -> lam mty mtm s0 s1 = lam mty mtm t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_tapp { mty mtm : nat } { s0 : tm mty mtm } { s1 : ty mty } { t0 : tm mty mtm } { t1 : ty mty } : s0 = t0 -> s1 = t1 -> tapp mty mtm s0 s1 = tapp mty mtm t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_tlam { mty mtm : nat } { s0 : tm (nat.succ mty) mtm } { t0 : tm (nat.succ mty) mtm } : s0 = t0 -> tlam mty mtm s0 = tlam mty mtm t0 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_app' { mty mtm : nat } { s0 : tm mty mtm } { s1 : tm mty mtm } { t0 : tm mty mtm } { t1 : tm mty mtm } (h0 : s0 = t0) (h1 : s1 = t1) : app mty mtm s0 s1 = app mty mtm t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_lam' { mty mtm : nat } { s0 : ty mty } { s1 : tm mty (nat.succ mtm) } { t0 : ty mty } { t1 : tm mty (nat.succ mtm) } (h0 : s0 = t0) (h1 : s1 = t1) : lam mty mtm s0 s1 = lam mty mtm t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_tapp' { mty mtm : nat } { s0 : tm mty mtm } { s1 : ty mty } { t0 : tm mty mtm } { t1 : ty mty } (h0 : s0 = t0) (h1 : s1 = t1) : tapp mty mtm s0 s1 = tapp mty mtm t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_tlam' { mty mtm : nat } { s0 : tm (nat.succ mty) mtm } { t0 : tm (nat.succ mty) mtm } (h0 : s0 = t0) : tlam mty mtm s0 = tlam mty mtm t0 :=
begin intros, congr; assumption end

@[substLemmas]
def upRen_ty_tm { m } { n } (xi : Fin m -> Fin n) : _ :=
  xi

@[substLemmas]
def upRen_tm_ty { m } { n } (xi : Fin m -> Fin n) : _ :=
  xi

@[substLemmas]
def upRen_tm_tm { m } { n } (xi : Fin m -> Fin n) : _ :=
  up_ren xi

@[substLemmas]
def ren_tm : Π { mty mtm : nat } { nty ntm : nat } (xity : Fin mty -> Fin nty) (xitm : Fin mtm -> Fin ntm) (s : tm mty mtm), tm nty ntm
    | mty mtm nty ntm xity xitm  (var_tm _ _ s) := (var_tm nty ntm) (xitm s)
    | mty mtm nty ntm xity xitm  (app _ _ s0 s1) := app nty ntm (ren_tm xity xitm s0) (ren_tm xity xitm s1)
    | mty mtm nty ntm xity xitm  (lam _ _ s0 s1) := lam nty ntm (ren_ty xity s0) (ren_tm (upRen_tm_ty xity) (upRen_tm_tm xitm) s1)
    | mty mtm nty ntm xity xitm  (tapp _ _ s0 s1) := tapp nty ntm (ren_tm xity xitm s0) (ren_ty xity s1)
    | mty mtm nty ntm xity xitm  (tlam _ _ s0) := tlam nty ntm (ren_tm (upRen_ty_ty xity) (upRen_ty_tm xitm) s0)

@[substLemmas]
def up_ty_tm { m } { nty ntm : nat } (sigma : Fin m -> tm nty ntm) : _ :=
  comp (ren_tm shift id) sigma

@[substLemmas]
def up_tm_ty { m } { nty : nat } (sigma : Fin m -> ty nty) : _ :=
  comp (ren_ty id) sigma

@[substLemmas]
def up_tm_tm { m } { nty ntm : nat } (sigma : Fin m -> tm nty ntm) : _ :=
  scons ((var_tm nty (nat.succ ntm)) var_zero) (comp (ren_tm id shift) sigma)

@[substLemmas]
def subst_tm : Π { mty mtm : nat } { nty ntm : nat } (sigmaty : Fin mty -> ty nty) (sigmatm : Fin mtm -> tm nty ntm) (s : tm mty mtm), tm nty ntm
    | mty mtm nty ntm sigmaty sigmatm  (var_tm _ _ s) := sigmatm s
    | mty mtm nty ntm sigmaty sigmatm  (app _ _ s0 s1) := app nty ntm (subst_tm sigmaty sigmatm s0) (subst_tm sigmaty sigmatm s1)
    | mty mtm nty ntm sigmaty sigmatm 
    (lam _ _ s0 s1) := lam nty ntm (subst_ty sigmaty s0) (subst_tm (up_tm_ty sigmaty) (up_tm_tm sigmatm) s1)
    | mty mtm nty ntm sigmaty sigmatm  (tapp _ _ s0 s1) := tapp nty ntm (subst_tm sigmaty sigmatm s0) (subst_ty sigmaty s1)
    | mty mtm nty ntm sigmaty sigmatm  (tlam _ _ s0) := tlam nty ntm (subst_tm (up_ty_ty sigmaty) (up_ty_tm sigmatm) s0)

@[substLemmas]
def upId_ty_tm { mty mtm : nat } { sigma : Fin mtm -> tm mty mtm } (Eq : forall x, sigma x = (var_tm mty mtm) x) : forall x, (up_ty_tm sigma) x = (var_tm (nat.succ mty) mtm) x :=
  λ n , have h : _, from ap (ren_tm shift id) (Eq n), begin simp with substLemmas at *, assumption end

@[substLemmas]
def upId_tm_ty { mty : nat } { sigma : Fin mty -> ty mty } (Eq : forall x, sigma x = (var_ty mty) x) : forall x, (up_tm_ty sigma) x = (var_ty mty) x :=
  λ n , have h : _, from ap (ren_ty id) (Eq n), begin simp with substLemmas at *, assumption end

@[substLemmas]
def upId_tm_tm { mty mtm : nat } { sigma : Fin mtm -> tm mty mtm } (Eq : forall x, sigma x = (var_tm mty mtm) x) : forall x, (up_tm_tm sigma) x = (var_tm mty (nat.succ mtm)) x :=
  λ n , match n with
  | (some n) := have h : _, from ap (ren_tm id shift) (Eq n), begin simp with substLemmas at *, assumption end
  | none := by refl
  end

@[substLemmas]
def idSubst_tm : Π { mty mtm : nat } (sigmaty : Fin mty -> ty mty) (sigmatm : Fin mtm -> tm mty mtm) (Eqty : forall x, sigmaty x = (var_ty mty) x) (Eqtm : forall x, sigmatm x = (var_tm mty mtm) x) (s : tm mty mtm), subst_tm sigmaty sigmatm s = s
    | mty mtm sigmaty sigmatm Eqty Eqtm  (var_tm _ _ s) := have h : _, from Eqtm s, begin simp with substLemmas at *, assumption end
    | mty mtm sigmaty sigmatm Eqty Eqtm 
    (app _ _ s0 s1) := have h : _, from congr_app (idSubst_tm sigmaty sigmatm Eqty Eqtm s0) (idSubst_tm sigmaty sigmatm Eqty Eqtm s1), begin simp with substLemmas at *, assumption end
    | mty mtm sigmaty sigmatm Eqty Eqtm 
    (lam _ _ s0 s1) := have h : _, from congr_lam (idSubst_ty sigmaty Eqty s0) (idSubst_tm (up_tm_ty sigmaty) (up_tm_tm sigmatm) (upId_tm_ty Eqty) (upId_tm_tm  Eqtm) s1), begin simp with substLemmas at *, assumption end
    | mty mtm sigmaty sigmatm Eqty Eqtm 
    (tapp _ _ s0 s1) := have h : _, from congr_tapp (idSubst_tm sigmaty sigmatm Eqty Eqtm s0) (idSubst_ty sigmaty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty mtm sigmaty sigmatm Eqty Eqtm 
    (tlam _ _ s0) := have h : _, from congr_tlam (idSubst_tm (up_ty_ty sigmaty) (up_ty_tm sigmatm) (upId_ty_ty  Eqty) (upId_ty_tm  Eqtm) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma idSubst'_tm { mty mtm : nat } { sigmaty : Fin mty -> ty mty } { sigmatm : Fin mtm -> tm mty mtm } { Eqty : forall x, sigmaty x = (var_ty mty) x } { Eqtm : forall x, sigmatm x = (var_tm mty mtm) x } { s s' : tm mty mtm } (Eqty : forall x, sigmaty x = (var_ty mty) x) (Eqtm : forall x, sigmatm x = (var_tm mty mtm) x) (Eqstm : s = s') : subst_tm sigmaty sigmatm s = s' :=
eq.trans (idSubst_tm sigmaty sigmatm Eqty Eqtm s) Eqstm

@[substLemmas]
def upExtRen_ty_tm { m } { n } (xi : Fin m -> Fin n) (zeta : Fin m -> Fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_ty_tm xi) x = (upRen_ty_tm zeta) x :=
  λ n , Eq n

@[substLemmas]
def upExtRen_tm_ty { m } { n } (xi : Fin m -> Fin n) (zeta : Fin m -> Fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_tm_ty xi) x = (upRen_tm_ty zeta) x :=
  λ n , Eq n

@[substLemmas]
def upExtRen_tm_tm { m } { n } (xi : Fin m -> Fin n) (zeta : Fin m -> Fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_tm_tm xi) x = (upRen_tm_tm zeta) x :=
  λ n , match n with
  | (some n) := ap shift (Eq n)
  | none := by refl
  end

@[substLemmas]
def extRen_tm : Π { mty mtm : nat } { nty ntm : nat } (xity : Fin mty -> Fin nty) (xitm : Fin mtm -> Fin ntm) (zetaty : Fin mty -> Fin nty) (zetatm : Fin mtm -> Fin ntm) (Eqty : forall x, xity x = zetaty x) (Eqtm : forall x, xitm x = zetatm x) (s : tm mty mtm), ren_tm xity xitm s = ren_tm zetaty zetatm s
    | mty mtm nty ntm xity xitm zetaty zetatm Eqty Eqtm 
    (var_tm _ _ s) := have h : _, from ap (var_tm nty ntm) (Eqtm s), begin simp with substLemmas at *, assumption end
    | mty mtm nty ntm xity xitm zetaty zetatm Eqty Eqtm 
    (app _ s0 s1) := have h : _, from congr_app (extRen_tm xity xitm zetaty zetatm Eqty Eqtm s0) (extRen_tm xity xitm zetaty zetatm Eqty Eqtm s1), begin simp with substLemmas at *, assumption end
    | mty mtm nty ntm xity xitm zetaty zetatm Eqty Eqtm 
    (lam _ s0 s1) := have h : _, from congr_lam (extRen_ty xity zetaty Eqty s0) (extRen_tm (upRen_tm_ty xity) (upRen_tm_tm xitm) (upRen_tm_ty zetaty) (upRen_tm_tm zetatm) (upExtRen_tm_ty xity zetaty Eqty) (upExtRen_tm_tm xitm zetatm Eqtm) s1), begin simp with substLemmas at *, assumption end
    | mty mtm nty ntm xity xitm zetaty zetatm Eqty Eqtm 
    (tapp _ s0 s1) := have h : _, from congr_tapp (extRen_tm xity xitm zetaty zetatm Eqty Eqtm s0) (extRen_ty xity zetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty mtm nty ntm xity xitm zetaty zetatm Eqty Eqtm 
    (tlam _ s0) := have h : _, from congr_tlam (extRen_tm (upRen_ty_ty xity) (upRen_ty_tm xitm) (upRen_ty_ty zetaty) (upRen_ty_tm zetatm) (upExtRen_ty_ty xity zetaty Eqty) (upExtRen_ty_tm xitm zetatm Eqtm) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma extRen'_tm { mty mtm : nat } { nty ntm : nat } { xity : Fin mty -> Fin nty } { xitm : Fin mtm -> Fin ntm } { zetaty : Fin mty -> Fin nty } { zetatm : Fin mtm -> Fin ntm } { s s' : tm mty mtm } (Eqty : forall x, xity x = zetaty x) (Eqtm : forall x, xitm x = zetatm x) (Eqstm : s = s') : ren_tm xity xitm s = ren_tm zetaty zetatm s' :=
have h : _, from extRen_tm xity xitm zetaty zetatm Eqty Eqtm s, begin rw <- Eqstm, assumption end

@[substLemmas]
def upExt_ty_tm { m } { nty ntm : nat } (sigma : Fin m -> tm nty ntm) (tau : Fin m -> tm nty ntm) (Eq : forall x, sigma x = tau x) : forall x, (up_ty_tm sigma) x = (up_ty_tm tau) x :=
  λ n , ap (ren_tm shift id) (Eq n)

@[substLemmas]
def upExt_tm_ty { m } { nty : nat } (sigma : Fin m -> ty nty) (tau : Fin m -> ty nty) (Eq : forall x, sigma x = tau x) : forall x, (up_tm_ty sigma) x = (up_tm_ty tau) x :=
  λ n , ap (ren_ty id) (Eq n)

@[substLemmas]
def upExt_tm_tm { m } { nty ntm : nat } (sigma : Fin m -> tm nty ntm) (tau : Fin m -> tm nty ntm) (Eq : forall x, sigma x = tau x) : forall x, (up_tm_tm sigma) x = (up_tm_tm tau) x :=
  λ n , match n with
  | (some n) := ap (ren_tm id shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
def ext_tm : Π { mty mtm : nat } { nty ntm : nat } (sigmaty : Fin mty -> ty nty) (sigmatm : Fin mtm -> tm nty ntm) (tauty : Fin mty -> ty nty) (tautm : Fin mtm -> tm nty ntm) (Eqty : forall x, sigmaty x = tauty x) (Eqtm : forall x, sigmatm x = tautm x) (s : tm mty mtm), subst_tm sigmaty sigmatm s = subst_tm tauty tautm s
    | mty mtm nty ntm sigmaty sigmatm tauty tautm Eqty Eqtm 
    (var_tm _ _ s) := have h : _, from Eqtm s, begin simp with substLemmas at *, assumption end
    | mty mtm nty ntm sigmaty sigmatm tauty tautm Eqty Eqtm 
    (app _ s0 s1) := have h : _, from congr_app (ext_tm sigmaty sigmatm tauty tautm Eqty Eqtm s0) (ext_tm sigmaty sigmatm tauty tautm Eqty Eqtm s1), begin simp with substLemmas at *, assumption end
    | mty mtm nty ntm sigmaty sigmatm tauty tautm Eqty Eqtm 
    (lam _ s0 s1) := have h : _, from congr_lam (ext_ty sigmaty tauty Eqty s0) (ext_tm (up_tm_ty sigmaty) (up_tm_tm sigmatm) (up_tm_ty tauty) (up_tm_tm tautm) (upExt_tm_ty sigmaty tauty Eqty) (upExt_tm_tm sigmatm tautm Eqtm) s1), begin simp with substLemmas at *, assumption end
    | mty mtm nty ntm sigmaty sigmatm tauty tautm Eqty Eqtm 
    (tapp _ s0 s1) := have h : _, from congr_tapp (ext_tm sigmaty sigmatm tauty tautm Eqty Eqtm s0) (ext_ty sigmaty tauty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty mtm nty ntm sigmaty sigmatm tauty tautm Eqty Eqtm 
    (tlam _ s0) := have h : _, from congr_tlam (ext_tm (up_ty_ty sigmaty) (up_ty_tm sigmatm) (up_ty_ty tauty) (up_ty_tm tautm) (upExt_ty_ty sigmaty tauty Eqty) (upExt_ty_tm sigmatm tautm Eqtm) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma ext'_tm { mty mtm : nat } { nty ntm : nat } { sigmaty : Fin mty -> ty nty } { sigmatm : Fin mtm -> tm nty ntm } { tauty : Fin mty -> ty nty } { tautm : Fin mtm -> tm nty ntm } { s s' : tm mty mtm } (Eqty : forall x, sigmaty x = tauty x) (Eqtm : forall x, sigmatm x = tautm x) (Eqstm : s = s') : subst_tm sigmaty sigmatm s = subst_tm tauty tautm s' :=
have h : _, from ext_tm sigmaty sigmatm tauty tautm Eqty Eqtm s, begin rw <- Eqstm, assumption end

@[substLemmas]
def compRenRen_tm : Π { kty ktm : nat } { lty ltm : nat } { mty mtm : nat } (xity : Fin mty -> Fin kty) (xitm : Fin mtm -> Fin ktm) (zetaty : Fin kty -> Fin lty) (zetatm : Fin ktm -> Fin ltm) (rhoty : Fin mty -> Fin lty) (rhotm : Fin mtm -> Fin ltm) (Eqty : forall x, (comp zetaty xity) x = rhoty x) (Eqtm : forall x, (comp zetatm xitm) x = rhotm x) (s : tm mty mtm), ren_tm zetaty zetatm (ren_tm xity xitm s) = ren_tm rhoty rhotm s
    | kty ktm lty ltm mty mtm xity xitm zetaty zetatm rhoty rhotm Eqty Eqtm 
    (var_tm _ _ s) := have h : _, from ap (var_tm lty ltm) (Eqtm s), begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm xity xitm zetaty zetatm rhoty rhotm Eqty Eqtm 
    (app _ s0 s1) := have h : _, from congr_app (compRenRen_tm xity xitm zetaty zetatm rhoty rhotm Eqty Eqtm s0) (compRenRen_tm xity xitm zetaty zetatm rhoty rhotm Eqty Eqtm s1), begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm xity xitm zetaty zetatm rhoty rhotm Eqty Eqtm 
    (lam _ s0 s1) := have h : _, from congr_lam (compRenRen_ty xity zetaty rhoty Eqty s0) (compRenRen_tm (upRen_tm_ty xity) (upRen_tm_tm xitm) (upRen_tm_ty zetaty) (upRen_tm_tm zetatm) (upRen_tm_ty rhoty) (upRen_tm_tm rhotm) Eqty (up_ren_ren xitm zetatm rhotm Eqtm) s1), begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm xity xitm zetaty zetatm rhoty rhotm Eqty Eqtm 
    (tapp _ s0 s1) := have h : _, from congr_tapp (compRenRen_tm xity xitm zetaty zetatm rhoty rhotm Eqty Eqtm s0) (compRenRen_ty xity zetaty rhoty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm xity xitm zetaty zetatm rhoty rhotm Eqty Eqtm 
    (tlam _ s0) := have h : _, from congr_tlam (compRenRen_tm (upRen_ty_ty xity) (upRen_ty_tm xitm) (upRen_ty_ty zetaty) (upRen_ty_tm zetatm) (upRen_ty_ty rhoty) (upRen_ty_tm rhotm) (up_ren_ren xity zetaty rhoty Eqty) Eqtm s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compRenRen'_tm { kty ktm : nat } { lty ltm : nat } { mty mtm : nat } { xity : Fin mty -> Fin kty } { xitm : Fin mtm -> Fin ktm } { zetaty : Fin kty -> Fin lty } { zetatm : Fin ktm -> Fin ltm } { rhoty : Fin mty -> Fin lty } { rhotm : Fin mtm -> Fin ltm } { s : tm mty mtm } { s' : tm kty ktm } (Eqty : forall x, (comp zetaty xity) x = rhoty x) (Eqtm : forall x, (comp zetatm xitm) x = rhotm x) (Eqstm : s' = ren_tm xity xitm s) : ren_tm zetaty zetatm s' = ren_tm rhoty rhotm s :=
have h : _, from compRenRen_tm xity xitm zetaty zetatm rhoty rhotm Eqty Eqtm s, begin rw Eqstm, assumption end

@[substLemmas]
def up_ren_subst_ty_tm { k } { l } { mty mtm : nat } (xi : Fin k -> Fin l) (tau : Fin l -> tm mty mtm) (theta : Fin k -> tm mty mtm) (Eq : forall x, (comp tau xi) x = theta x) : forall x, (comp (up_ty_tm tau) (upRen_ty_tm xi)) x = (up_ty_tm theta) x :=
  λ n , ap (ren_tm shift id) (Eq n)

@[substLemmas]
def up_ren_subst_tm_ty { k } { l } { mty : nat } (xi : Fin k -> Fin l) (tau : Fin l -> ty mty) (theta : Fin k -> ty mty) (Eq : forall x, (comp tau xi) x = theta x) : forall x, (comp (up_tm_ty tau) (upRen_tm_ty xi)) x = (up_tm_ty theta) x :=
  λ n , ap (ren_ty id) (Eq n)

@[substLemmas]
def up_ren_subst_tm_tm { k } { l } { mty mtm : nat } (xi : Fin k -> Fin l) (tau : Fin l -> tm mty mtm) (theta : Fin k -> tm mty mtm) (Eq : forall x, (comp tau xi) x = theta x) : forall x, (comp (up_tm_tm tau) (upRen_tm_tm xi)) x = (up_tm_tm theta) x :=
  λ n , match n with
  | (some n) := ap (ren_tm id shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
def compRenSubst_tm : Π { kty ktm : nat } { lty ltm : nat } { mty mtm : nat } (xity : Fin mty -> Fin kty) (xitm : Fin mtm -> Fin ktm) (tauty : Fin kty -> ty lty) (tautm : Fin ktm -> tm lty ltm) (thetaty : Fin mty -> ty lty) (thetatm : Fin mtm -> tm lty ltm) (Eqty : forall x, (comp tauty xity) x = thetaty x) (Eqtm : forall x, (comp tautm xitm) x = thetatm x) (s : tm mty mtm), subst_tm tauty tautm (ren_tm xity xitm s) = subst_tm thetaty thetatm s
    | kty ktm lty ltm mty mtm xity xitm tauty tautm thetaty thetatm Eqty Eqtm 
    (var_tm _ _ s) := have h : _, from Eqtm s, begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm xity xitm tauty tautm thetaty thetatm Eqty Eqtm 
    (app _ s0 s1) := have h : _, from congr_app (compRenSubst_tm xity xitm tauty tautm thetaty thetatm Eqty Eqtm s0) (compRenSubst_tm xity xitm tauty tautm thetaty thetatm Eqty Eqtm s1), begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm xity xitm tauty tautm thetaty thetatm Eqty Eqtm 
    (lam _ s0 s1) := have h : _, from congr_lam (compRenSubst_ty xity tauty thetaty Eqty s0) (compRenSubst_tm (upRen_tm_ty xity) (upRen_tm_tm xitm) (up_tm_ty tauty) (up_tm_tm tautm) (up_tm_ty thetaty) (up_tm_tm thetatm) (up_ren_subst_tm_ty xity tauty thetaty Eqty) (up_ren_subst_tm_tm xitm tautm thetatm Eqtm) s1), begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm xity xitm tauty tautm thetaty thetatm Eqty Eqtm 
    (tapp _ s0 s1) := have h : _, from congr_tapp (compRenSubst_tm xity xitm tauty tautm thetaty thetatm Eqty Eqtm s0) (compRenSubst_ty xity tauty thetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm xity xitm tauty tautm thetaty thetatm Eqty Eqtm 
    (tlam _ s0) := have h : _, from congr_tlam (compRenSubst_tm (upRen_ty_ty xity) (upRen_ty_tm xitm) (up_ty_ty tauty) (up_ty_tm tautm) (up_ty_ty thetaty) (up_ty_tm thetatm) (up_ren_subst_ty_ty xity tauty thetaty Eqty) (up_ren_subst_ty_tm xitm tautm thetatm Eqtm) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compRenSubst'_tm { kty ktm : nat } { lty ltm : nat } { mty mtm : nat } { xity : Fin mty -> Fin kty } { xitm : Fin mtm -> Fin ktm } { tauty : Fin kty -> ty lty } { tautm : Fin ktm -> tm lty ltm } { thetaty : Fin mty -> ty lty } { thetatm : Fin mtm -> tm lty ltm } { s : tm mty mtm } { s' : tm kty ktm } (Eqty : forall x, (comp tauty xity) x = thetaty x) (Eqtm : forall x, (comp tautm xitm) x = thetatm x) (Eqstm : s' = ren_tm xity xitm s) : subst_tm tauty tautm s' = subst_tm thetaty thetatm s :=
have h : _, from compRenSubst_tm xity xitm tauty tautm thetaty thetatm Eqty Eqtm s, begin rw Eqstm, assumption end

@[substLemmas]
def up_subst_ren_ty_tm { k } { lty ltm : nat } { mty mtm : nat } (sigma : Fin k -> tm lty ltm) (zetaty : Fin lty -> Fin mty) (zetatm : Fin ltm -> Fin mtm) (theta : Fin k -> tm mty mtm) (Eq : forall x, (comp (ren_tm zetaty zetatm) sigma) x = theta x) : forall x, (comp (ren_tm (upRen_ty_ty zetaty) (upRen_ty_tm zetatm)) (up_ty_tm sigma)) x = (up_ty_tm theta) x :=
  λ n , eq.trans (compRenRen_tm shift id (upRen_ty_ty zetaty) (upRen_ty_tm zetatm) (comp shift zetaty) (comp id zetatm) (λ x , by refl) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compRenRen_tm zetaty zetatm shift id (comp shift zetaty) (comp id zetatm) (λ x , by refl) (λ x , by refl) (sigma n))) (ap (ren_tm shift id) (Eq n)))

@[substLemmas]
def up_subst_ren_tm_ty { k } { lty : nat } { mty : nat } (sigma : Fin k -> ty lty) (zetaty : Fin lty -> Fin mty) (theta : Fin k -> ty mty) (Eq : forall x, (comp (ren_ty zetaty) sigma) x = theta x) : forall x, (comp (ren_ty (upRen_tm_ty zetaty)) (up_tm_ty sigma)) x = (up_tm_ty theta) x :=
  λ n , eq.trans (compRenRen_ty id (upRen_tm_ty zetaty) (comp id zetaty) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compRenRen_ty zetaty id (comp id zetaty) (λ x , by refl) (sigma n))) (ap (ren_ty id) (Eq n)))

@[substLemmas]
def up_subst_ren_tm_tm { k } { lty ltm : nat } { mty mtm : nat } (sigma : Fin k -> tm lty ltm) (zetaty : Fin lty -> Fin mty) (zetatm : Fin ltm -> Fin mtm) (theta : Fin k -> tm mty mtm) (Eq : forall x, (comp (ren_tm zetaty zetatm) sigma) x = theta x) : forall x, (comp (ren_tm (upRen_tm_ty zetaty) (upRen_tm_tm zetatm)) (up_tm_tm sigma)) x = (up_tm_tm theta) x :=
  λ n , match n with
  | (some n) := eq.trans (compRenRen_tm id shift (upRen_tm_ty zetaty) (upRen_tm_tm zetatm) (comp id zetaty) (comp shift zetatm) (λ x , by refl) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compRenRen_tm zetaty zetatm id shift (comp id zetaty) (comp shift zetatm) (λ x , by refl) (λ x , by refl) (sigma n))) (ap (ren_tm id shift) (Eq n)))
  | none := by refl
  end

@[substLemmas]
def compSubstRen_tm : Π { kty ktm : nat } { lty ltm : nat } { mty mtm : nat } (sigmaty : Fin mty -> ty kty) (sigmatm : Fin mtm -> tm kty ktm) (zetaty : Fin kty -> Fin lty) (zetatm : Fin ktm -> Fin ltm) (thetaty : Fin mty -> ty lty) (thetatm : Fin mtm -> tm lty ltm) (Eqty : forall x, (comp (ren_ty zetaty) sigmaty) x = thetaty x) (Eqtm : forall x, (comp (ren_tm zetaty zetatm) sigmatm) x = thetatm x) (s : tm mty mtm), ren_tm zetaty zetatm (subst_tm sigmaty sigmatm s) = subst_tm thetaty thetatm s
    | kty ktm lty ltm mty mtm sigmaty sigmatm zetaty zetatm thetaty thetatm Eqty Eqtm 
    (var_tm _ _ s) := have h : _, from Eqtm s, begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm sigmaty sigmatm zetaty zetatm thetaty thetatm Eqty Eqtm 
    (app _ s0 s1) := have h : _, from congr_app (compSubstRen_tm sigmaty sigmatm zetaty zetatm thetaty thetatm Eqty Eqtm s0) (compSubstRen_tm sigmaty sigmatm zetaty zetatm thetaty thetatm Eqty Eqtm s1), begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm sigmaty sigmatm zetaty zetatm thetaty thetatm Eqty Eqtm 
    (lam _ s0 s1) := have h : _, from congr_lam (compSubstRen_ty sigmaty zetaty thetaty Eqty s0) (compSubstRen_tm (up_tm_ty sigmaty) (up_tm_tm sigmatm) (upRen_tm_ty zetaty) (upRen_tm_tm zetatm) (up_tm_ty thetaty) (up_tm_tm thetatm) (up_subst_ren_tm_ty sigmaty zetaty thetaty Eqty) (up_subst_ren_tm_tm sigmatm zetaty zetatm thetatm Eqtm) s1), begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm sigmaty sigmatm zetaty zetatm thetaty thetatm Eqty Eqtm 
    (tapp _ s0 s1) := have h : _, from congr_tapp (compSubstRen_tm sigmaty sigmatm zetaty zetatm thetaty thetatm Eqty Eqtm s0) (compSubstRen_ty sigmaty zetaty thetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm sigmaty sigmatm zetaty zetatm thetaty thetatm Eqty Eqtm 
    (tlam _ s0) := have h : _, from congr_tlam (compSubstRen_tm (up_ty_ty sigmaty) (up_ty_tm sigmatm) (upRen_ty_ty zetaty) (upRen_ty_tm zetatm) (up_ty_ty thetaty) (up_ty_tm thetatm) (up_subst_ren_ty_ty sigmaty zetaty thetaty Eqty) (up_subst_ren_ty_tm sigmatm zetaty zetatm thetatm Eqtm) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compSubstRen'_tm { kty ktm : nat } { lty ltm : nat } { mty mtm : nat } { sigmaty : Fin mty -> ty kty } { sigmatm : Fin mtm -> tm kty ktm } { zetaty : Fin kty -> Fin lty } { zetatm : Fin ktm -> Fin ltm } { thetaty : Fin mty -> ty lty } { thetatm : Fin mtm -> tm lty ltm } { s : tm mty mtm } { s' : tm kty ktm } (Eqty : forall x, (comp (ren_ty zetaty) sigmaty) x = thetaty x) (Eqtm : forall x, (comp (ren_tm zetaty zetatm) sigmatm) x = thetatm x) (Eqstm : s' = subst_tm sigmaty sigmatm s) : ren_tm zetaty zetatm s' = subst_tm thetaty thetatm s :=
have h : _, from compSubstRen_tm sigmaty sigmatm zetaty zetatm thetaty thetatm Eqty Eqtm s, begin rw Eqstm, assumption end

@[substLemmas]
def up_subst_subst_ty_tm { k } { lty ltm : nat } { mty mtm : nat } (sigma : Fin k -> tm lty ltm) (tauty : Fin lty -> ty mty) (tautm : Fin ltm -> tm mty mtm) (theta : Fin k -> tm mty mtm) (Eq : forall x, (comp (subst_tm tauty tautm) sigma) x = theta x) : forall x, (comp (subst_tm (up_ty_ty tauty) (up_ty_tm tautm)) (up_ty_tm sigma)) x = (up_ty_tm theta) x :=
  λ n , eq.trans (compRenSubst_tm shift id (up_ty_ty tauty) (up_ty_tm tautm) (comp (up_ty_ty tauty) shift) (comp (up_ty_tm tautm) id) (λ x , by refl) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compSubstRen_tm tauty tautm shift id (comp (ren_ty shift) tauty) (comp (ren_tm shift id) tautm) (λ x , by refl) (λ x , by refl) (sigma n))) (ap (ren_tm shift id) (Eq n)))

@[substLemmas]
def up_subst_subst_tm_ty { k } { lty : nat } { mty : nat } (sigma : Fin k -> ty lty) (tauty : Fin lty -> ty mty) (theta : Fin k -> ty mty) (Eq : forall x, (comp (subst_ty tauty) sigma) x = theta x) : forall x, (comp (subst_ty (up_tm_ty tauty)) (up_tm_ty sigma)) x = (up_tm_ty theta) x :=
  λ n , eq.trans (compRenSubst_ty id (up_tm_ty tauty) (comp (up_tm_ty tauty) id) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compSubstRen_ty tauty id (comp (ren_ty id) tauty) (λ x , by refl) (sigma n))) (ap (ren_ty id) (Eq n)))

@[substLemmas]
def up_subst_subst_tm_tm { k } { lty ltm : nat } { mty mtm : nat } (sigma : Fin k -> tm lty ltm) (tauty : Fin lty -> ty mty) (tautm : Fin ltm -> tm mty mtm) (theta : Fin k -> tm mty mtm) (Eq : forall x, (comp (subst_tm tauty tautm) sigma) x = theta x) : forall x, (comp (subst_tm (up_tm_ty tauty) (up_tm_tm tautm)) (up_tm_tm sigma)) x = (up_tm_tm theta) x :=
  λ n , match n with
  | (some n) := eq.trans (compRenSubst_tm id shift (up_tm_ty tauty) (up_tm_tm tautm) (comp (up_tm_ty tauty) id) (comp (up_tm_tm tautm) shift) (λ x , by refl) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compSubstRen_tm tauty tautm id shift (comp (ren_ty id) tauty) (comp (ren_tm id shift) tautm) (λ x , by refl) (λ x , by refl) (sigma n))) (ap (ren_tm id shift) (Eq n)))
  | none := by refl
  end

@[substLemmas]
def compSubstSubst_tm : Π { kty ktm : nat } { lty ltm : nat } { mty mtm : nat } (sigmaty : Fin mty -> ty kty) (sigmatm : Fin mtm -> tm kty ktm) (tauty : Fin kty -> ty lty) (tautm : Fin ktm -> tm lty ltm) (thetaty : Fin mty -> ty lty) (thetatm : Fin mtm -> tm lty ltm) (Eqty : forall x, (comp (subst_ty tauty) sigmaty) x = thetaty x) (Eqtm : forall x, (comp (subst_tm tauty tautm) sigmatm) x = thetatm x) (s : tm mty mtm), subst_tm tauty tautm (subst_tm sigmaty sigmatm s) = subst_tm thetaty thetatm s
    | kty ktm lty ltm mty mtm sigmaty sigmatm tauty tautm thetaty thetatm Eqty Eqtm 
    (var_tm _ _ s) := have h : _, from Eqtm s, begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm sigmaty sigmatm tauty tautm thetaty thetatm Eqty Eqtm 
    (app _ s0 s1) := have h : _, from congr_app (compSubstSubst_tm sigmaty sigmatm tauty tautm thetaty thetatm Eqty Eqtm s0) (compSubstSubst_tm sigmaty sigmatm tauty tautm thetaty thetatm Eqty Eqtm s1), begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm sigmaty sigmatm tauty tautm thetaty thetatm Eqty Eqtm 
    (lam _ s0 s1) := have h : _, from congr_lam (compSubstSubst_ty sigmaty tauty thetaty Eqty s0) (compSubstSubst_tm (up_tm_ty sigmaty) (up_tm_tm sigmatm) (up_tm_ty tauty) (up_tm_tm tautm) (up_tm_ty thetaty) (up_tm_tm thetatm) (up_subst_subst_tm_ty sigmaty tauty thetaty Eqty) (up_subst_subst_tm_tm sigmatm tauty tautm thetatm Eqtm) s1), begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm sigmaty sigmatm tauty tautm thetaty thetatm Eqty Eqtm 
    (tapp _ s0 s1) := have h : _, from congr_tapp (compSubstSubst_tm sigmaty sigmatm tauty tautm thetaty thetatm Eqty Eqtm s0) (compSubstSubst_ty sigmaty tauty thetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty ktm lty ltm mty mtm sigmaty sigmatm tauty tautm thetaty thetatm Eqty Eqtm 
    (tlam _ s0) := have h : _, from congr_tlam (compSubstSubst_tm (up_ty_ty sigmaty) (up_ty_tm sigmatm) (up_ty_ty tauty) (up_ty_tm tautm) (up_ty_ty thetaty) (up_ty_tm thetatm) (up_subst_subst_ty_ty sigmaty tauty thetaty Eqty) (up_subst_subst_ty_tm sigmatm tauty tautm thetatm Eqtm) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compSubstSubst'_tm { kty ktm : nat } { lty ltm : nat } { mty mtm : nat } { sigmaty : Fin mty -> ty kty } { sigmatm : Fin mtm -> tm kty ktm } { tauty : Fin kty -> ty lty } { tautm : Fin ktm -> tm lty ltm } { thetaty : Fin mty -> ty lty } { thetatm : Fin mtm -> tm lty ltm } { s : tm mty mtm } { s' : tm kty ktm } (Eqty : forall x, (comp (subst_ty tauty) sigmaty) x = thetaty x) (Eqtm : forall x, (comp (subst_tm tauty tautm) sigmatm) x = thetatm x) (Eqstm : s' = subst_tm sigmaty sigmatm s) : subst_tm tauty tautm s' = subst_tm thetaty thetatm s :=
have h : _, from compSubstSubst_tm sigmaty sigmatm tauty tautm thetaty thetatm Eqty Eqtm s, begin rw Eqstm, assumption end

@[substLemmas]
lemma comp'_tm { mty mtm : nat } { nty ntm : nat } { sigma'ty : Fin mty -> ty nty } { sigma'tm : Fin mtm -> tm nty ntm } { tau'ty : Fin mty -> ty nty } { tau'tm : Fin mtm -> tm nty ntm } (Eqsty : forall x, sigmaty x = sigma'ty x) (Eqstm : forall x, sigmatm x = sigma'tm x) (Eqtty : forall x, sigmaty x = sigma'ty x) (Eqttm : forall x, sigmatm x = sigma'tm x) : forall x, (comp (subst_tm tauty tautm) sigmatm) x = (comp (subst_tm tau'ty tau'tm) sigmatm) x :=
begin assumption end

@[substLemmas]
def rinstInst_up_ty_tm { m } { nty ntm : nat } (xi : Fin m -> Fin ntm) (sigma : Fin m -> tm nty ntm) (Eq : forall x, (comp (var_tm nty ntm) xi) x = sigma x) : forall x, (comp (var_tm (nat.succ nty) ntm) (upRen_ty_tm xi)) x = (up_ty_tm sigma) x :=
  λ n , ap (ren_tm shift id) (Eq n)

@[substLemmas]
def rinstInst_up_tm_ty { m } { nty : nat } (xi : Fin m -> Fin nty) (sigma : Fin m -> ty nty) (Eq : forall x, (comp (var_ty nty) xi) x = sigma x) : forall x, (comp (var_ty nty) (upRen_tm_ty xi)) x = (up_tm_ty sigma) x :=
  λ n , ap (ren_ty id) (Eq n)

@[substLemmas]
def rinstInst_up_tm_tm { m } { nty ntm : nat } (xi : Fin m -> Fin ntm) (sigma : Fin m -> tm nty ntm) (Eq : forall x, (comp (var_tm nty ntm) xi) x = sigma x) : forall x, (comp (var_tm nty (nat.succ ntm)) (upRen_tm_tm xi)) x = (up_tm_tm sigma) x :=
  λ n , match n with
  | (some n) := ap (ren_tm id shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
def rinst_inst_tm : Π { mty mtm : nat } { nty ntm : nat } (xity : Fin mty -> Fin nty) (xitm : Fin mtm -> Fin ntm) (sigmaty : Fin mty -> ty nty) (sigmatm : Fin mtm -> tm nty ntm) (Eqty : forall x, (comp (var_ty nty) xity) x = sigmaty x) (Eqtm : forall x, (comp (var_tm nty ntm) xitm) x = sigmatm x) (s : tm mty mtm), ren_tm xity xitm s = subst_tm sigmaty sigmatm s
    | mty mtm nty ntm xity xitm sigmaty sigmatm Eqty Eqtm 
    (var_tm _ _ s) := have h : _, from Eqtm s, begin simp with substLemmas at *, assumption end
    | mty mtm nty ntm xity xitm sigmaty sigmatm Eqty Eqtm 
    (app _ s0 s1) := have h : _, from congr_app (rinst_inst_tm xity xitm sigmaty sigmatm Eqty Eqtm s0) (rinst_inst_tm xity xitm sigmaty sigmatm Eqty Eqtm s1), begin simp with substLemmas at *, assumption end
    | mty mtm nty ntm xity xitm sigmaty sigmatm Eqty Eqtm 
    (lam _ s0 s1) := have h : _, from congr_lam (rinst_inst_ty xity sigmaty Eqty s0) (rinst_inst_tm (upRen_tm_ty xity) (upRen_tm_tm xitm) (up_tm_ty sigmaty) (up_tm_tm sigmatm) (rinstInst_up_tm_ty xity sigmaty Eqty) (rinstInst_up_tm_tm xitm sigmatm Eqtm) s1), begin simp with substLemmas at *, assumption end
    | mty mtm nty ntm xity xitm sigmaty sigmatm Eqty Eqtm 
    (tapp _ s0 s1) := have h : _, from congr_tapp (rinst_inst_tm xity xitm sigmaty sigmatm Eqty Eqtm s0) (rinst_inst_ty xity sigmaty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty mtm nty ntm xity xitm sigmaty sigmatm Eqty Eqtm 
    (tlam _ s0) := have h : _, from congr_tlam (rinst_inst_tm (upRen_ty_ty xity) (upRen_ty_tm xitm) (up_ty_ty sigmaty) (up_ty_tm sigmatm) (rinstInst_up_ty_ty xity sigmaty Eqty) (rinstInst_up_ty_tm xitm sigmatm Eqtm) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma rinst_inst'_tm { mty mtm : nat } { nty ntm : nat } { sigmaty : Fin mty -> ty nty } { sigmatm : Fin mtm -> tm nty ntm } { xity : Fin mty -> Fin nty } { xitm : Fin mtm -> Fin ntm } { s s' : tm mty mtm } (Eqty : forall x, (comp (var_ty nty) xity) x = sigmaty x) (Eqtm : forall x, (comp (var_tm nty ntm) xitm) x = sigmatm x) (Eqstm : s = s') : ren_tm xity xitm s' = subst_tm sigmaty sigmatm s :=
have h : _, from rinst_inst_tm xity xitm sigmaty sigmatm Eqty Eqtm s, begin rw <- Eqstm, assumption end

@[substLemmas]
lemma instId_tm { mty mtm : nat } : subst_tm (var_ty mty) (var_tm mty mtm) = id :=
 (λ x , idSubst_tm (var_ty mty) (var_tm mty mtm) (λ n , by refl) (λ n , by refl) (id x))

@[substLemmas]
lemma varL_tm { mty mtm : nat } { nty ntm : nat } (sigmaty : Fin mty -> ty nty) (sigmatm : Fin mtm -> tm nty ntm) : comp (subst_tm sigmaty sigmatm) (var_tm mty mtm) = sigmatm :=
 (λ x , by refl)

@[substLemmas]
lemma rinstInst_tm { mty mtm : nat } { nty ntm : nat } (xity : Fin mty -> Fin nty) (xitm : Fin mtm -> Fin ntm) : ren_tm xity xitm = subst_tm (comp (var_ty nty) xity) (comp (var_tm nty ntm) xitm) :=
 (λ x , rinst_inst_tm xity xitm (_) (_) (λ n , by refl) (λ n , by refl) x)

@[substLemmas]
lemma compComp_tm { kty ktm : nat } { lty ltm : nat } { mty mtm : nat } (sigmaty : Fin mty -> ty kty) (sigmatm : Fin mtm -> tm kty ktm) (tauty : Fin kty -> ty lty) (tautm : Fin ktm -> tm lty ltm) (s : tm mty mtm) : subst_tm tauty tautm (subst_tm sigmaty sigmatm s) = subst_tm (comp (subst_ty tauty) sigmaty) (comp (subst_tm tauty tautm) sigmatm) s :=
compSubstSubst_tm sigmaty sigmatm tauty tautm (_) (_) (λ n , by refl) (λ n , by refl) s

@[substLemmas]
lemma compComp'_tm { kty ktm : nat } { lty ltm : nat } { mty mtm : nat } (sigmaty : Fin mty -> ty kty) (sigmatm : Fin mtm -> tm kty ktm) (tauty : Fin kty -> ty lty) (tautm : Fin ktm -> tm lty ltm) : comp (subst_tm tauty tautm) (subst_tm sigmaty sigmatm) = subst_tm (comp (subst_ty tauty) sigmaty) (comp (subst_tm tauty tautm) sigmatm) :=
 (λ n , compComp_tm sigmaty sigmatm tauty tautm n)

meta def is_app_tm : expr -> bool
  | `(app _ _ %%s0 %%s1) := tt
  | `(congr_app' %%s0 %%s1) := tt
  | `(idSubst'_tm %%Eqty %%Eqtm %%Eqstm) := is_app_tm Eqstm
  | _ := ff

meta def destruct_app_tm : expr -> tactic (expr × expr)
  | `(app _ _ %%s0 %%s1) := return (s0, s1)
  | `(congr_app' %%s0 %%s1) := return (s0, s1)
  | `(idSubst'_tm %%Eqty %%Eqtm %%Eqstm) := destruct_app_tm Eqstm
  | e := return (e, e)

meta def is_lam_tm : expr -> bool
  | `(lam _ _ %%s0 %%s1) := tt
  | `(congr_lam' %%s0 %%s1) := tt
  | `(idSubst'_tm %%Eqty %%Eqtm %%Eqstm) := is_lam_tm Eqstm
  | _ := ff

meta def destruct_lam_tm : expr -> tactic (expr × expr)
  | `(lam _ _ %%s0 %%s1) := return (s0, s1)
  | `(congr_lam' %%s0 %%s1) := return (s0, s1)
  | `(idSubst'_tm %%Eqty %%Eqtm %%Eqstm) := destruct_lam_tm Eqstm
  | e := return (e, e)

meta def is_tapp_tm : expr -> bool
  | `(tapp _ _ %%s0 %%s1) := tt
  | `(congr_tapp' %%s0 %%s1) := tt
  | `(idSubst'_tm %%Eqty %%Eqtm %%Eqstm) := is_tapp_tm Eqstm
  | _ := ff

meta def destruct_tapp_tm : expr -> tactic (expr × expr)
  | `(tapp _ _ %%s0 %%s1) := return (s0, s1)
  | `(congr_tapp' %%s0 %%s1) := return (s0, s1)
  | `(idSubst'_tm %%Eqty %%Eqtm %%Eqstm) := destruct_tapp_tm Eqstm
  | e := return (e, e)

meta def is_tlam_tm : expr -> bool
  | `(tlam _ _ %%s0) := tt
  | `(congr_tlam' %%s0) := tt
  | `(idSubst'_tm %%Eqty %%Eqtm %%Eqstm) := is_tlam_tm Eqstm
  | _ := ff

meta def destruct_tlam_tm : expr -> tactic expr
  | `(tlam _ _ %%s0) := return s0
  | `(congr_tlam' %%s0) := return s0
  | `(idSubst'_tm %%Eqty %%Eqtm %%Eqstm) := destruct_tlam_tm Eqstm
  | e := return e

meta def is_ren_tm : expr -> bool
  | `(extRen'_tm %%Eqty %%Eqtm %%Eqstm) := tt
  | _ := ff

meta def destruct_ren_tm : expr -> tactic (expr × expr)
  | `(extRen'_tm %%Eqty %%Eqtm %%Eqstm) := return (Eqty, Eqtm, Eqstm)
  | `(%%e) := return (e,e)

meta def is_subst_tm : expr -> bool
  | `(ext'_tm %%Eqty %%Eqtm %%Eqstm) := tt
  | _ := ff

meta def destruct_subst_tm : expr -> tactic (expr × expr)
  | `(ext'_tm %%Eqty %%Eqtm %%Eqstm) := return (Eqty, Eqtm, Eqstm)
  | `(%%e) := return (e,e)

meta def is_InstRen_tm : expr -> bool
  | `(compRenRen'_tm %%Eqty %%Eqtm %%Eqstm) := is_InstRen_tm Eqstm
  | _ := ff

meta def destruct_InstRen_tm : expr -> tactic (expr × expr)
  | `(compRenRen'_tm %%Eqty %%Eqtm %%Eqstm) := destruct_InstRen_tm Eqstm
  | `(%%e) := return (e,e)

meta def is_Inst_tm : expr -> bool
  | `(compRenRen'_tm %%Eqty %%Eqtm %%Eqstm) := is_Inst_tm Eqstm
  | `(compRenSubst'_tm %%Eqty %%Eqtm %%Eqstm) := is_Inst_tm Eqstm
  | `(compSubstRen'_tm %%Eqty %%Eqtm %%Eqstm) := is_Inst_tm Eqstm
  | `(compSubstSubst'_tm %%Eqty %%Eqtm %%Eqstm) := is_Inst_tm Eqstm
  | `(idSubst'_tm %%Eqty %%Eqtm %%Eqstm) := is_Inst_tm Eqstm
  | `(rinst_inst'_tm %%Eqty %%Eqtm %%Eqstm) := is_Inst_tm Eqstm
  | _ := ff

meta def destruct_Inst_tm : expr -> tactic (expr × expr)
  | `(ext'_tm %%Eqty %%Eqtm %%Eqstm) := return (Eqty, Eqtm, Eqstm)
  | `(compRenRen'_tm %%Eqty %%Eqtm %%Eqstm) := destruct_Inst_tm Eqstm
  | `(compRenSubst'_tm %%Eqty %%Eqtm %%Eqstm) := destruct_Inst_tm Eqstm
  | `(compSubstRen'_tm %%Eqty %%Eqtm %%Eqstm) := destruct_Inst_tm Eqstm
  | `(compSubstSubst'_tm %%Eqty %%Eqtm %%Eqstm) := destruct_Inst_tm Eqstm
  | `(idSubst'_tm %%Eqty %%Eqtm %%Eqstm) := destruct_Inst_tm Eqstm
  | `(rinst_inst'_tm %%Eqty %%Eqtm %%Eqstm) := destruct_Inst_tm Eqstm
  | `(%%e) := return (e,e)

meta def is_scons : expr -> bool
  | `(up_ren_subst_ty_tm %%xi %%tau %%theta %%Eq) := is_scons Eq
  | `(up_ren_subst_tm_ty %%xi %%tau %%theta %%Eq) := is_scons Eq
  | `(up_ren_subst_tm_tm %%xi %%tau %%theta %%Eq) := is_scons Eq
  | `(up_subst_ren_ty_tm %%sigma %%zetaty %%zetatm %%theta %%Eq) := is_scons Eq
  | `(up_subst_ren_tm_ty %%sigma %%zetaty %%theta %%Eq) := is_scons Eq
  | `(up_subst_ren_tm_tm %%sigma %%zetaty %%zetatm %%theta %%Eq) := is_scons Eq
  | `(up_subst_subst_ty_tm %%sigma %%tauty %%tautm %%theta %%Eq) := is_scons Eq
  | `(up_subst_subst_tm_ty %%sigma %%tauty %%theta %%Eq) := is_scons Eq
  | `(up_subst_subst_tm_tm %%sigma %%tauty %%tautm %%theta %%Eq) := is_scons Eq
  | `(scons %%H1 %%H2) := tt
  | _ := ff

meta def is_comp : expr -> bool
  | `(up_ren_subst_ty_tm %%xi %%tau %%theta %%Eq) := is_comp Eq
  | `(up_ren_subst_tm_ty %%xi %%tau %%theta %%Eq) := is_comp Eq
  | `(up_ren_subst_tm_tm %%xi %%tau %%theta %%Eq) := is_comp Eq
  | `(up_subst_ren_ty_tm %%sigma %%zetaty %%zetatm %%theta %%Eq) := is_comp Eq
  | `(up_subst_ren_tm_ty %%sigma %%zetaty %%theta %%Eq) := is_comp Eq
  | `(up_subst_ren_tm_tm %%sigma %%zetaty %%zetatm %%theta %%Eq) := is_comp Eq
  | `(up_subst_subst_ty_tm %%sigma %%tauty %%tautm %%theta %%Eq) := is_comp Eq
  | `(up_subst_subst_tm_ty %%sigma %%tauty %%theta %%Eq) := is_comp Eq
  | `(up_subst_subst_tm_tm %%sigma %%tauty %%tautm %%theta %%Eq) := is_comp Eq
  | `(comp %%H1 %%H2) := tt
  | _ := ff

meta def is_var_tm : expr -> bool
  | `(up_ren_subst_ty_tm %%xi %%tau %%theta %%Eq) := is_var_tm Eq
  | `(up_ren_subst_tm_ty %%xi %%tau %%theta %%Eq) := is_var_tm Eq
  | `(up_ren_subst_tm_tm %%xi %%tau %%theta %%Eq) := is_var_tm Eq
  | `(up_subst_ren_ty_tm %%sigma %%zetaty %%zetatm %%theta %%Eq) := is_var_tm Eq
  | `(up_subst_ren_tm_ty %%sigma %%zetaty %%theta %%Eq) := is_var_tm Eq
  | `(up_subst_ren_tm_tm %%sigma %%zetaty %%zetatm %%theta %%Eq) := is_var_tm Eq
  | `(up_subst_subst_ty_tm %%sigma %%tauty %%tautm %%theta %%Eq) := is_var_tm Eq
  | `(up_subst_subst_tm_ty %%sigma %%tauty %%theta %%Eq) := is_var_tm Eq
  | `(up_subst_subst_tm_tm %%sigma %%tauty %%tautm %%theta %%Eq) := is_var_tm Eq
  | `(var_tm %%H1 %%H2) := tt
  | _ := ff

meta def destruct_scons : expr -> tactic (expr × expr)
  | `(up_ren_subst_ty_tm %%xi %%tau %%theta %%Eq) := destruct_scons Eq
  | `(up_ren_subst_tm_ty %%xi %%tau %%theta %%Eq) := destruct_scons Eq
  | `(up_ren_subst_tm_tm %%xi %%tau %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_ren_ty_tm %%sigma %%zetaty %%zetatm %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_ren_tm_ty %%sigma %%zetaty %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_ren_tm_tm %%sigma %%zetaty %%zetatm %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_subst_ty_tm %%sigma %%tauty %%tautm %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_subst_tm_ty %%sigma %%tauty %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_subst_tm_tm %%sigma %%tauty %%tautm %%theta %%Eq) := destruct_scons Eq
  | `(scons %%H1 %%H2) := return (H1, H2)
  | `(%%e) := return (e,e)

meta def destruct_comp : expr -> tactic (expr × expr)
  | `(up_ren_subst_ty_tm %%xi %%tau %%theta %%Eq) := destruct_comp Eq
  | `(up_ren_subst_tm_ty %%xi %%tau %%theta %%Eq) := destruct_comp Eq
  | `(up_ren_subst_tm_tm %%xi %%tau %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_ren_ty_tm %%sigma %%zetaty %%zetatm %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_ren_tm_ty %%sigma %%zetaty %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_ren_tm_tm %%sigma %%zetaty %%zetatm %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_subst_ty_tm %%sigma %%tauty %%tautm %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_subst_tm_ty %%sigma %%tauty %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_subst_tm_tm %%sigma %%tauty %%tautm %%theta %%Eq) := destruct_comp Eq
  | `(comp %%H1 %%H2) := return (H1, H2)
  | `(%%e) := return (e,e)

meta def destruct_var_tm : expr -> tactic expr
  | `(up_ren_subst_ty_tm %%xi %%tau %%theta %%Eq) := destruct_var_tm Eq
  | `(up_ren_subst_tm_ty %%xi %%tau %%theta %%Eq) := destruct_var_tm Eq
  | `(up_ren_subst_tm_tm %%xi %%tau %%theta %%Eq) := destruct_var_tm Eq
  | `(up_subst_ren_ty_tm %%sigma %%zetaty %%zetatm %%theta %%Eq) := destruct_var_tm Eq
  | `(up_subst_ren_tm_ty %%sigma %%zetaty %%theta %%Eq) := destruct_var_tm Eq
  | `(up_subst_ren_tm_tm %%sigma %%zetaty %%zetatm %%theta %%Eq) := destruct_var_tm Eq
  | `(up_subst_subst_ty_tm %%sigma %%tauty %%tautm %%theta %%Eq) := destruct_var_tm Eq
  | `(up_subst_subst_tm_ty %%sigma %%tauty %%theta %%Eq) := destruct_var_tm Eq
  | `(up_subst_subst_tm_tm %%sigma %%tauty %%tautm %%theta %%Eq) := destruct_var_tm Eq
  | `(var_tm %%s) := return s
  | `(%%e) := return e

meta def normalize'_tm : nat -> expr -> expr -> tactic expr
  | n e1 e2 := if (n = Inst)
  then if (is_var_tm e2)
  then ex ``(idSubst'_tm %%e1 %%e2)
  else if (is_lam_tm e1)
  then do e3 ← destruct_lam_tm e1, e4 ← normalize'_tm Up e2 e2, e5 ← normalize'_tm Inst e3 e1, ex ``(congr_lam' %%e4 %%e5)
  else if (is_subst_tm e2)
  then do (e3,e4) ← destruct_subst_tm e2, e5 ← normalize'_tm Comp e4 e2, e6 ← normalize'_tm Inst e3 e5, ex ``(compSubstSubst'_tm %%e5 %%e1 %%e6)
  else ex ``(ext'_tm %%e1 %%e2)
  else if (n = Rinst)
  then if (is_var_tm e2)
  then ex ``(idSubst'_tm %%e1 %%e2)
  else if (is_lam_tm e1)
  then do e3 ← destruct_lam_tm e1, e4 ← normalize'_tm Up e2 e2, e5 ← normalize'_tm Rinst e3 e1, ex ``(congr_lam' %%e4 %%e5)
  else if (is_ren_tm e2)
  then do (e3,e4) ← destruct_ren_tm e2, e5 ← normalize'_tm Comp e4 e2, e6 ← normalize'_tm Rinst e3 e5, ex ``(compRenRen'_tm %%e5 %%e1 %%e6)
  else ex ``(extRen'_tm %%e1 %%e2)
  else if (n = Comp)
  then if (is_scons e1)
  then do (e3,e4) ← destruct_scons e1, e5 ← normalize'_tm Inst e3 e2, e6 ← normalize'_tm Comp e4 e2, e7 ← normalize'_tm Cons e5 e6, ex ``(scons_comp' %%e5 %%e1 %%e6)
  else ex ``(comp'_tm %%e1 %%e2)
  else if (n = Up)
  then TODO
  else ex ``(%%e1=%%e2)
