import .fintype
import .wftactic
run_cmd mk_simp_attr `substLemmas

@[substLemmas]
inductive tm  : nat -> Type 
  | var_tm : Π { ntm : nat }, Fin ntm -> tm ntm
  | app : Π { ntm : nat }, tm ntm -> tm ntm -> tm ntm
  | lam : Π { ntm : nat }, tm (nat.succ ntm) -> tm ntm
open tm 

@[substLemmas]
lemma congr_app { mtm : nat } { s0 : tm mtm } { s1 : tm mtm } { t0 : tm mtm } { t1 : tm mtm } : s0 = t0 -> s1 = t1 -> app s0 s1 = app t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_lam { mtm : nat } { s0 : tm (nat.succ mtm) } { t0 : tm (nat.succ mtm) } : s0 = t0 -> lam s0 = lam t0 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_app' { mtm : nat } { s0 : tm mtm } { s1 : tm mtm } { t0 : tm mtm } { t1 : tm mtm } (h0 : s0 = t0) (h1 : s1 = t1) : app s0 s1 = app t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_lam' { mtm : nat } { s0 : tm (nat.succ mtm) } { t0 : tm (nat.succ mtm) } (h0 : s0 = t0) : lam s0 = lam t0 :=
begin intros, congr; assumption end

@[substLemmas]
def upRen_tm_tm { m } { n } (xi : Fin m -> Fin n) : _ :=
  up_ren xi

@[substLemmas]
def ren_tm : Π { mtm : nat } { ntm : nat } (xitm : Fin mtm -> Fin ntm) (s : tm mtm), tm ntm
    | mtm ntm xitm  (var_tm s) := var_tm (xitm s)
    | mtm ntm xitm  (app s0 s1) := app (ren_tm xitm s0) (ren_tm xitm s1)
    | mtm ntm xitm  (lam s0) := lam (ren_tm (upRen_tm_tm xitm) s0)

@[substLemmas]
def up_tm_tm { m } { ntm : nat } (sigma : Fin m -> tm ntm) : _ :=
  scons (var_tm var_zero) (comp (ren_tm shift) sigma)

@[substLemmas]
def subst_tm : Π { mtm : nat } { ntm : nat } (sigmatm : Fin mtm -> tm ntm) (s : tm mtm), tm ntm
    | mtm ntm sigmatm  (var_tm s) := sigmatm s
    | mtm ntm sigmatm  (app s0 s1) := app (subst_tm sigmatm s0) (subst_tm sigmatm s1)
    | mtm ntm sigmatm  (lam s0) := lam (subst_tm (up_tm_tm sigmatm) s0)

@[substLemmas]
def upId_tm_tm { mtm : nat } { sigma : Fin mtm -> tm mtm } (Eq : forall x, sigma x = var_tm x) : forall x, (up_tm_tm sigma) x = var_tm x :=
  λ n , match n with
  | (some n) := have h : _, from ap (ren_tm shift) (Eq n), begin simp with substLemmas at *, assumption end
  | none := by refl
  end

@[substLemmas]
def idSubst_tm : Π { mtm : nat } (sigmatm : Fin mtm -> tm mtm) (Eqtm : forall x, sigmatm x = (var_tm mtm) x) (s : tm mtm), subst_tm sigmatm s = s
    | mtm sigmatm Eqtm  (var_tm s) := have h : _, from Eqtm s, begin simp with substLemmas at *, assumption end
    | mtm sigmatm Eqtm 
    (app s0 s1) := have h : _, from congr_app (idSubst_tm sigmatm Eqtm s0) (idSubst_tm sigmatm Eqtm s1), begin simp with substLemmas at *, assumption end
    | mtm sigmatm Eqtm 
    (lam s0) := have h : _, from congr_lam (idSubst_tm (up_tm_tm sigmatm) (upId_tm_tm  Eqtm) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma idSubst'_tm { mtm : nat } { sigmatm : Fin mtm -> tm mtm } { Eqtm : forall x, sigmatm x = (var_tm mtm) x } { s s' : tm mtm } (Eqtm : forall x, sigmatm x = (var_tm mtm) x) (Eqstm : s = s') : subst_tm sigmatm s = s' :=
eq.trans (idSubst_tm sigmatm Eqtm s) Eqstm

@[substLemmas]
def upExtRen_tm_tm { m } { n } (xi : Fin m -> Fin n) (zeta : Fin m -> Fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_tm_tm xi) x = (upRen_tm_tm zeta) x :=
  λ n , match n with
  | (some n) := ap shift (Eq n)
  | none := by refl
  end

@[substLemmas]
def extRen_tm : Π { mtm : nat } { ntm : nat } (xitm : Fin mtm -> Fin ntm) (zetatm : Fin mtm -> Fin ntm) (Eqtm : forall x, xitm x = zetatm x) (s : tm mtm), ren_tm xitm s = ren_tm zetatm s
    | mtm ntm xitm zetatm Eqtm  (var_tm s) := have h : _, from ap var_tm (Eqtm s), begin simp with substLemmas at *, assumption end
    | mtm ntm xitm zetatm Eqtm 
    (app s0 s1) := have h : _, from congr_app (extRen_tm xitm zetatm Eqtm s0) (extRen_tm xitm zetatm Eqtm s1), begin simp with substLemmas at *, assumption end
    | mtm ntm xitm zetatm Eqtm 
    (lam s0) := have h : _, from congr_lam (extRen_tm (upRen_tm_tm xitm) (upRen_tm_tm zetatm) (upExtRen_tm_tm xitm zetatm Eqtm) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma extRen'_tm { mtm : nat } { ntm : nat } { xitm : Fin mtm -> Fin ntm } { zetatm : Fin mtm -> Fin ntm } { s s' : tm mtm } (Eqtm : forall x, xitm x = zetatm x) (Eqstm : s = s') : ren_tm xitm s = ren_tm zetatm s' :=
have h : _, from extRen_tm xitm zetatm Eqtm s, begin rw <- Eqstm, assumption end

@[substLemmas]
def upExt_tm_tm { m } { ntm : nat } (sigma : Fin m -> tm ntm) (tau : Fin m -> tm ntm) (Eq : forall x, sigma x = tau x) : forall x, (up_tm_tm sigma) x = (up_tm_tm tau) x :=
  λ n , match n with
  | (some n) := ap (ren_tm shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
def ext_tm : Π { mtm : nat } { ntm : nat } (sigmatm : Fin mtm -> tm ntm) (tautm : Fin mtm -> tm ntm) (Eqtm : forall x, sigmatm x = tautm x) (s : tm mtm), subst_tm sigmatm s = subst_tm tautm s
    | mtm ntm sigmatm tautm Eqtm  (var_tm s) := have h : _, from Eqtm s, begin simp with substLemmas at *, assumption end
    | mtm ntm sigmatm tautm Eqtm 
    (app s0 s1) := have h : _, from congr_app (ext_tm sigmatm tautm Eqtm s0) (ext_tm sigmatm tautm Eqtm s1), begin simp with substLemmas at *, assumption end
    | mtm ntm sigmatm tautm Eqtm 
    (lam s0) := have h : _, from congr_lam (ext_tm (up_tm_tm sigmatm) (up_tm_tm tautm) (upExt_tm_tm sigmatm tautm Eqtm) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma ext'_tm { mtm : nat } { ntm : nat } { sigmatm : Fin mtm -> tm ntm } { tautm : Fin mtm -> tm ntm } { s s' : tm mtm } (Eqtm : forall x, sigmatm x = tautm x) (Eqstm : s = s') : subst_tm sigmatm s = subst_tm tautm s' :=
have h : _, from ext_tm sigmatm tautm Eqtm s, begin rw <- Eqstm, assumption end

@[substLemmas]
def compRenRen_tm : Π { ktm : nat } { ltm : nat } { mtm : nat } (xitm : Fin mtm -> Fin ktm) (zetatm : Fin ktm -> Fin ltm) (rhotm : Fin mtm -> Fin ltm) (Eqtm : forall x, (comp zetatm xitm) x = rhotm x) (s : tm mtm), ren_tm zetatm (ren_tm xitm s) = ren_tm rhotm s
    | ktm ltm mtm xitm zetatm rhotm Eqtm 
    (var_tm s) := have h : _, from ap var_tm (Eqtm s), begin simp with substLemmas at *, assumption end
    | ktm ltm mtm xitm zetatm rhotm Eqtm 
    (app s0 s1) := have h : _, from congr_app (compRenRen_tm xitm zetatm rhotm Eqtm s0) (compRenRen_tm xitm zetatm rhotm Eqtm s1), begin simp with substLemmas at *, assumption end
    | ktm ltm mtm xitm zetatm rhotm Eqtm 
    (lam s0) := have h : _, from congr_lam (compRenRen_tm (upRen_tm_tm xitm) (upRen_tm_tm zetatm) (upRen_tm_tm rhotm) (up_ren_ren xitm zetatm rhotm Eqtm) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compRenRen'_tm { ktm : nat } { ltm : nat } { mtm : nat } { xitm : Fin mtm -> Fin ktm } { zetatm : Fin ktm -> Fin ltm } { rhotm : Fin mtm -> Fin ltm } { s : tm mtm } { s' : tm ktm } (Eqtm : forall x, (comp zetatm xitm) x = rhotm x) (Eqstm : s' = ren_tm xitm s) : ren_tm zetatm s' = ren_tm rhotm s :=
have h : _, from compRenRen_tm xitm zetatm rhotm Eqtm s, begin rw Eqstm, assumption end

@[substLemmas]
def up_ren_subst_tm_tm { k } { l } { mtm : nat } (xi : Fin k -> Fin l) (tau : Fin l -> tm mtm) (theta : Fin k -> tm mtm) (Eq : forall x, (comp tau xi) x = theta x) : forall x, (comp (up_tm_tm tau) (upRen_tm_tm xi)) x = (up_tm_tm theta) x :=
  λ n , match n with
  | (some n) := ap (ren_tm shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
def compRenSubst_tm : Π { ktm : nat } { ltm : nat } { mtm : nat } (xitm : Fin mtm -> Fin ktm) (tautm : Fin ktm -> tm ltm) (thetatm : Fin mtm -> tm ltm) (Eqtm : forall x, (comp tautm xitm) x = thetatm x) (s : tm mtm), subst_tm tautm (ren_tm xitm s) = subst_tm thetatm s
    | ktm ltm mtm xitm tautm thetatm Eqtm  (var_tm s) := have h : _, from Eqtm s, begin simp with substLemmas at *, assumption end
    | ktm ltm mtm xitm tautm thetatm Eqtm 
    (app s0 s1) := have h : _, from congr_app (compRenSubst_tm xitm tautm thetatm Eqtm s0) (compRenSubst_tm xitm tautm thetatm Eqtm s1), begin simp with substLemmas at *, assumption end
    | ktm ltm mtm xitm tautm thetatm Eqtm 
    (lam s0) := have h : _, from congr_lam (compRenSubst_tm (upRen_tm_tm xitm) (up_tm_tm tautm) (up_tm_tm thetatm) (up_ren_subst_tm_tm xitm tautm thetatm Eqtm) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compRenSubst'_tm { ktm : nat } { ltm : nat } { mtm : nat } { xitm : Fin mtm -> Fin ktm } { tautm : Fin ktm -> tm ltm } { thetatm : Fin mtm -> tm ltm } { s : tm mtm } { s' : tm ktm } (Eqtm : forall x, (comp tautm xitm) x = thetatm x) (Eqstm : s' = ren_tm xitm s) : subst_tm tautm s' = subst_tm thetatm s :=
have h : _, from compRenSubst_tm xitm tautm thetatm Eqtm s, begin rw Eqstm, assumption end

@[substLemmas]
def up_subst_ren_tm_tm { k } { ltm : nat } { mtm : nat } (sigma : Fin k -> tm ltm) (zetatm : Fin ltm -> Fin mtm) (theta : Fin k -> tm mtm) (Eq : forall x, (comp (ren_tm zetatm) sigma) x = theta x) : forall x, (comp (ren_tm (upRen_tm_tm zetatm)) (up_tm_tm sigma)) x = (up_tm_tm theta) x :=
  λ n , match n with
  | (some n) := eq.trans (compRenRen_tm shift (upRen_tm_tm zetatm) (comp shift zetatm) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compRenRen_tm zetatm shift (comp shift zetatm) (λ x , by refl) (sigma n))) (ap (ren_tm shift) (Eq n)))
  | none := by refl
  end

@[substLemmas]
def compSubstRen_tm : Π { ktm : nat } { ltm : nat } { mtm : nat } (sigmatm : Fin mtm -> tm ktm) (zetatm : Fin ktm -> Fin ltm) (thetatm : Fin mtm -> tm ltm) (Eqtm : forall x, (comp (ren_tm zetatm) sigmatm) x = thetatm x) (s : tm mtm), ren_tm zetatm (subst_tm sigmatm s) = subst_tm thetatm s
    | ktm ltm mtm sigmatm zetatm thetatm Eqtm  (var_tm s) := have h : _, from Eqtm s, begin simp with substLemmas at *, assumption end
    | ktm ltm mtm sigmatm zetatm thetatm Eqtm 
    (app s0 s1) := have h : _, from congr_app (compSubstRen_tm sigmatm zetatm thetatm Eqtm s0) (compSubstRen_tm sigmatm zetatm thetatm Eqtm s1), begin simp with substLemmas at *, assumption end
    | ktm ltm mtm sigmatm zetatm thetatm Eqtm 
    (lam s0) := have h : _, from congr_lam (compSubstRen_tm (up_tm_tm sigmatm) (upRen_tm_tm zetatm) (up_tm_tm thetatm) (up_subst_ren_tm_tm sigmatm zetatm thetatm Eqtm) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compSubstRen'_tm { ktm : nat } { ltm : nat } { mtm : nat } { sigmatm : Fin mtm -> tm ktm } { zetatm : Fin ktm -> Fin ltm } { thetatm : Fin mtm -> tm ltm } { s : tm mtm } { s' : tm ktm } (Eqtm : forall x, (comp (ren_tm zetatm) sigmatm) x = thetatm x) (Eqstm : s' = subst_tm sigmatm s) : ren_tm zetatm s' = subst_tm thetatm s :=
have h : _, from compSubstRen_tm sigmatm zetatm thetatm Eqtm s, begin rw Eqstm, assumption end

@[substLemmas]
def up_subst_subst_tm_tm { k } { ltm : nat } { mtm : nat } (sigma : Fin k -> tm ltm) (tautm : Fin ltm -> tm mtm) (theta : Fin k -> tm mtm) (Eq : forall x, (comp (subst_tm tautm) sigma) x = theta x) : forall x, (comp (subst_tm (up_tm_tm tautm)) (up_tm_tm sigma)) x = (up_tm_tm theta) x :=
  λ n , match n with
  | (some n) := eq.trans (compRenSubst_tm shift (up_tm_tm tautm) (comp (up_tm_tm tautm) shift) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compSubstRen_tm tautm shift (comp (ren_tm shift) tautm) (λ x , by refl) (sigma n))) (ap (ren_tm shift) (Eq n)))
  | none := by refl
  end

@[substLemmas]
def compSubstSubst_tm : Π { ktm : nat } { ltm : nat } { mtm : nat } (sigmatm : Fin mtm -> tm ktm) (tautm : Fin ktm -> tm ltm) (thetatm : Fin mtm -> tm ltm) (Eqtm : forall x, (comp (subst_tm tautm) sigmatm) x = thetatm x) (s : tm mtm), subst_tm tautm (subst_tm sigmatm s) = subst_tm thetatm s
    | ktm ltm mtm sigmatm tautm thetatm Eqtm  (var_tm s) := have h : _, from Eqtm s, begin simp with substLemmas at *, assumption end
    | ktm ltm mtm sigmatm tautm thetatm Eqtm 
    (app s0 s1) := have h : _, from congr_app (compSubstSubst_tm sigmatm tautm thetatm Eqtm s0) (compSubstSubst_tm sigmatm tautm thetatm Eqtm s1), begin simp with substLemmas at *, assumption end
    | ktm ltm mtm sigmatm tautm thetatm Eqtm 
    (lam s0) := have h : _, from congr_lam (compSubstSubst_tm (up_tm_tm sigmatm) (up_tm_tm tautm) (up_tm_tm thetatm) (up_subst_subst_tm_tm sigmatm tautm thetatm Eqtm) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compSubstSubst'_tm { ktm : nat } { ltm : nat } { mtm : nat } { sigmatm : Fin mtm -> tm ktm } { tautm : Fin ktm -> tm ltm } { thetatm : Fin mtm -> tm ltm } { s : tm mtm } { s' : tm ktm } (Eqtm : forall x, (comp (subst_tm tautm) sigmatm) x = thetatm x) (Eqstm : s' = subst_tm sigmatm s) : subst_tm tautm s' = subst_tm thetatm s :=
have h : _, from compSubstSubst_tm sigmatm tautm thetatm Eqtm s, begin rw Eqstm, assumption end

@[substLemmas]
def rinstInst_up_tm_tm { m } { ntm : nat } (xi : Fin m -> Fin ntm) (sigma : Fin m -> tm ntm) (Eq : forall x, (comp var_tm xi) x = sigma x) : forall x, (comp var_tm (upRen_tm_tm xi)) x = (up_tm_tm sigma) x :=
  λ n , match n with
  | (some n) := have h : _, from ap (ren_tm shift) (Eq n), begin simp with substLemmas at *, assumption end
  | none := by refl
  end

@[substLemmas]
def rinst_inst_tm : Π { mtm : nat } { ntm : nat } (xitm : Fin mtm -> Fin ntm) (sigmatm : Fin mtm -> tm ntm) (Eqtm : forall x, (comp var_tm xitm) x = sigmatm x) (s : tm mtm), ren_tm xitm s = subst_tm sigmatm s
    | mtm ntm xitm sigmatm Eqtm  (var_tm s) := have h : _, from Eqtm s, begin simp with substLemmas at *, assumption end
    | mtm ntm xitm sigmatm Eqtm 
    (app s0 s1) := have h : _, from congr_app (rinst_inst_tm xitm sigmatm Eqtm s0) (rinst_inst_tm xitm sigmatm Eqtm s1), begin simp with substLemmas at *, assumption end
    | mtm ntm xitm sigmatm Eqtm 
    (lam s0) := have h : _, from congr_lam (rinst_inst_tm (upRen_tm_tm xitm) (up_tm_tm sigmatm) (rinstInst_up_tm_tm xitm sigmatm Eqtm) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma rinst_inst'_tm { mtm : nat } { ntm : nat } { sigmatm : Fin mtm -> tm ntm } { xitm : Fin mtm -> Fin ntm } { s s' : tm mtm } (Eqtm : forall x, (comp var_tm xitm) x = sigmatm x) (Eqstm : s = s') : ren_tm xitm s' = subst_tm sigmatm s :=
have h : _, from rinst_inst_tm xitm sigmatm Eqtm s, begin rw <- Eqstm, assumption end

@[substLemmas]
lemma rinstInst_tm { mtm : nat } { ntm : nat } (xitm : Fin mtm -> Fin ntm) : ren_tm xitm = subst_tm (comp var_tm xitm) :=
funext (λ x , rinst_inst_tm xitm (_) (λ n , by refl) x)

@[substLemmas]
lemma instId_tm { mtm : nat } : subst_tm (@var_tm mtm) = id := --TODO
funext (λ x , idSubst_tm (var_tm) (λ n , by refl) (id x))

@[substLemmas]
lemma rinstId_tm { mtm : nat } : @ren_tm mtm mtm id = id :=
eq.trans (rinstInst_tm id) instId_tm

@[substLemmas]
lemma varL_tm { mtm : nat } { ntm : nat } (sigmatm : Fin mtm -> tm ntm) : comp (subst_tm sigmatm) var_tm = sigmatm :=
funext (λ x , by refl)

@[substLemmas]
lemma varLRen_tm { mtm : nat } { ntm : nat } (xitm : Fin mtm -> Fin ntm) : comp (ren_tm xitm) var_tm = comp var_tm xitm :=
funext (λ x , by refl)

@[substLemmas]
lemma compComp_tm { ktm : nat } { ltm : nat } { mtm : nat } (sigmatm : Fin mtm -> tm ktm) (tautm : Fin ktm -> tm ltm) (s : tm mtm) : subst_tm tautm (subst_tm sigmatm s) = subst_tm (comp (subst_tm tautm) sigmatm) s :=
compSubstSubst_tm sigmatm tautm (_) (λ n , by refl) s

@[substLemmas]
lemma compComp'_tm { ktm : nat } { ltm : nat } { mtm : nat } (sigmatm : Fin mtm -> tm ktm) (tautm : Fin ktm -> tm ltm) : comp (subst_tm tautm) (subst_tm sigmatm) = subst_tm (comp (subst_tm tautm) sigmatm) :=
funext (λ n , compComp_tm sigmatm tautm n)

@[substLemmas]
lemma compRen_tm { ktm : nat } { ltm : nat } { mtm : nat } (sigmatm : Fin mtm -> tm ktm) (xitm : Fin ktm -> Fin ltm) (s : tm mtm) : ren_tm xitm (subst_tm sigmatm s) = subst_tm (comp (ren_tm xitm) sigmatm) s :=
compSubstRen_tm sigmatm xitm (_) (λ n , by refl) s

@[substLemmas]
lemma compRen'_tm { ktm : nat } { ltm : nat } { mtm : nat } (sigmatm : Fin mtm -> tm ktm) (xitm : Fin ktm -> Fin ltm) : comp (ren_tm xitm) (subst_tm sigmatm) = subst_tm (comp (ren_tm xitm) sigmatm) :=
funext (λ n , compRen_tm sigmatm xitm n)

@[substLemmas]
lemma renComp_tm { ktm : nat } { ltm : nat } { mtm : nat } (xitm : Fin mtm -> Fin ktm) (sigmatm : Fin ktm -> tm ltm) (s : tm mtm) : subst_tm sigmatm (ren_tm xitm s) = subst_tm (comp sigmatm xitm) s :=
compRenSubst_tm xitm sigmatm (_) (λ n , by refl) s

@[substLemmas]
lemma renComp'_tm { ktm : nat } { ltm : nat } { mtm : nat } (xitm : Fin mtm -> Fin ktm) (sigmatm : Fin ktm -> tm ltm) : comp (subst_tm sigmatm) (ren_tm xitm) = subst_tm (comp sigmatm xitm) :=
funext (λ n , renComp_tm xitm sigmatm n)

@[substLemmas]
lemma renRen_tm { ktm : nat } { ltm : nat } { mtm : nat } (xitm : Fin mtm -> Fin ktm) (rhotm : Fin ktm -> Fin ltm) (s : tm mtm) : ren_tm rhotm (ren_tm xitm s) = ren_tm (comp rhotm xitm) s :=
compRenRen_tm xitm rhotm (_) (λ n , by refl) s

@[substLemmas]
lemma renRen'_tm { ktm : nat } { ltm : nat } { mtm : nat } (xitm : Fin mtm -> Fin ktm) (rhotm : Fin ktm -> Fin ltm) : comp (ren_tm rhotm) (ren_tm xitm) = ren_tm (comp rhotm xitm) :=
funext (λ n , renRen_tm xitm rhotm n)
#check compComp_tm
#check compComp'_tm
#check compSubstSubst'_tm
#check compSubstSubst_tm
meta def is_app_tm : expr -> bool
  | `(app _ %%s0 %%s1) := tt
  | `(congr_app' %%s0 %%s1) := tt
  | `(idSubst'_tm %%Eqtm %%Eqstm) := is_app_tm Eqstm
  | _ := ff

meta def destruct_app_tm : expr -> tactic (expr × expr)
  | `(app _ %%s0 %%s1) := return (s0, s1)
  | `(congr_app' %%s0 %%s1) := return (s0, s1)
  | `(idSubst'_tm %%Eqtm %%Eqstm) := destruct_app_tm Eqstm
  | e := return (e, e)

meta def is_lam_tm : expr -> bool
  | `(lam _ %%s0) := tt
  | `(congr_lam' %%s0) := tt
  | `(idSubst'_tm %%Eqtm %%Eqstm) := is_lam_tm Eqstm
  | _ := ff

meta def destruct_lam_tm : expr -> tactic expr
  | `(lam _ %%s0) := return s0
  | `(congr_lam' %%s0) := return s0
  | `(idSubst'_tm %%Eqtm %%Eqstm) := destruct_lam_tm Eqstm
  | e := return e

meta def is_ren_tm : expr -> bool
  | `(ren_tm %%xi %%s) := tt
  | `(extRen'_tm %%Eqtm %%Eqstm) := tt
  | _ := ff

meta def destruct_ren_tm : expr -> tactic (expr × expr)
  | `(ren_tm %%xi %%s) := return (xi, s)
  | `(extRen'_tm %%Eqtm %%Eqstm) := return (Eqtm, Eqstm)
  | `(%%e) := return (e,e)

meta def is_subst_tm : expr -> bool
  | `(subst_tm %%xi %%s) := tt
  | `(ren_tm %%xi %%s) := tt
  | `(ext'_tm %%Eqtm %%Eqstm) := return (Eqtm, Eqstm)
  | _ := ff

meta def destruct_subst_tm : expr -> tactic (expr × expr)
  | `(subst_tm %%xi %%s) := return (xi, s)
  | `(ren_tm %%xi %%s) := return (xi, s)
  | `(ext'_tm %%Eqtm %%Eqstm) := return (Eqtm, Eqstm)
  | `(%%e) := return (e,e)

meta def is_InstRen_tm : expr -> bool
  | `(compRenRen'_tm %%Eqtm %%Eqstm) := is_InstRen_tm Eqstm
  | e := is_ren_tm e

meta def destruct_InstRen_tm : expr -> tactic (expr × expr)
  | `(compRenRen'_tm %%Eqtm %%Eqstm) := destruct_InstRen_tm Eqstm
  | e := destruct_ren_tm e

meta def is_Inst_tm : expr -> bool
  | `(compRenRen'_tm %%Eqtm %%Eqstm) := is_Inst_tm Eqstm
  | `(compRenSubst'_tm %%Eqtm %%Eqstm) := is_Inst_tm Eqstm
  | `(compSubstRen'_tm %%Eqtm %%Eqstm) := is_Inst_tm Eqstm
  | `(compSubstSubst'_tm %%Eqtm %%Eqstm) := is_Inst_tm Eqstm
  | `(idSubst'_tm %%Eqtm %%Eqstm) := is_Inst_tm Eqstm
  | `(rinst_inst'_tm %%Eqtm %%Eqstm) := is_Inst_tm Eqstm
  | e := is_subst_tm e

meta def destruct_Inst_tm : expr -> tactic (expr × expr)
  | `(ext'_tm %%Eqtm %%Eqstm) := return (Eqtm, Eqstm)
  | `(compRenRen'_tm %%Eqtm %%Eqstm) := destruct_Inst_tm Eqstm
  | `(compRenSubst'_tm %%Eqtm %%Eqstm) := destruct_Inst_tm Eqstm
  | `(compSubstRen'_tm %%Eqtm %%Eqstm) := destruct_Inst_tm Eqstm
  | `(compSubstSubst'_tm %%Eqtm %%Eqstm) := destruct_Inst_tm Eqstm
  | `(idSubst'_tm %%Eqtm %%Eqstm) := destruct_Inst_tm Eqstm
  | `(rinst_inst'_tm %%Eqtm %%Eqstm) := destruct_Inst_tm Eqstm
  | e := destruct_subst_tm e

meta def is_scons : expr -> bool
  | `(up_ren_subst_tm_tm %%xi %%tau %%theta %%Eq) := is_scons Eq
  | `(up_subst_ren_tm_tm %%sigma %%zetatm %%theta %%Eq) := is_scons Eq
  | `(up_subst_subst_tm_tm %%sigma %%tautm %%theta %%Eq) := is_scons Eq
  | `(scons %%H1 %%H2) := tt
  | _ := ff

meta def is_comp : expr -> bool
  | `(up_ren_subst_tm_tm %%xi %%tau %%theta %%Eq) := is_comp Eq
  | `(up_subst_ren_tm_tm %%sigma %%zetatm %%theta %%Eq) := is_comp Eq
  | `(up_subst_subst_tm_tm %%sigma %%tautm %%theta %%Eq) := is_comp Eq
  | `(comp %%H1 %%H2) := tt
  | _ := ff

meta def is_var_tm : expr -> bool
  | `(up_ren_subst_tm_tm %%xi %%tau %%theta %%Eq) := is_var_tm Eq
  | `(up_subst_ren_tm_tm %%sigma %%zetatm %%theta %%Eq) := is_var_tm Eq
  | `(up_subst_subst_tm_tm %%sigma %%tautm %%theta %%Eq) := is_var_tm Eq
  | `(var_tm %%H1 %%H2) := tt
  | _ := ff

meta def destruct_scons : expr -> tactic (expr × expr)
  | `(up_ren_subst_tm_tm %%xi %%tau %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_ren_tm_tm %%sigma %%zetatm %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_subst_tm_tm %%sigma %%tautm %%theta %%Eq) := destruct_scons Eq
  | `(scons %%H1 %%H2) := return (H1, H2)
  | `(%%e) := return (e,e)

meta def destruct_comp : expr -> tactic (expr × expr)
  | `(up_ren_subst_tm_tm %%xi %%tau %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_ren_tm_tm %%sigma %%zetatm %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_subst_tm_tm %%sigma %%tautm %%theta %%Eq) := destruct_comp Eq
  | `(comp %%H1 %%H2) := return (H1, H2)
  | `(%%e) := return (e,e)

meta def destruct_var_tm : expr -> tactic expr
  | `(up_ren_subst_tm_tm %%xi %%tau %%theta %%Eq) := destruct_var_tm Eq
  | `(up_subst_ren_tm_tm %%sigma %%zetatm %%theta %%Eq) := destruct_var_tm Eq
  | `(up_subst_subst_tm_tm %%sigma %%tautm %%theta %%Eq) := destruct_var_tm Eq
  | `(var_tm %%s) := return s
  | `(%%e) := return e

meta def normalize_tm : nat -> expr -> expr -> tactic expr
  | n e1 e2 := if (n = Inst)
  then if (is_var_tm e1)
  then ex ``(idSubst'_tm %%e1 %%e2)
  
  else if (is_lam_tm e2)
  then do e3 ← destruct_lam_tm e2, e4 ← normalize_tm Up e1 e1, e5 ← normalize_tm Inst e3 e2, ex ``(congr_lam' %%e4 %%e5)
  
  else if (is_subst_tm e2)
  then do (e3,e4) ← destruct_subst_tm e2, e5 ← normalize_tm Comp e4 e1, e6 ← normalize_tm Inst e3 e5, ex ``(compSubstSubst'_tm %%e5 %%e2 %%e6)
  
  else ex ``(ext'_tm %%e1 %%e2)
  
  else if (n = Rinst)
  then if (is_var_tm e1)
  then ex ``(idSubst'_tm %%e1 %%e2)
  
  else if (is_lam_tm e2)
  then do e3 ← destruct_lam_tm e2, e4 ← normalize_tm Up e1 e1, e5 ← normalize_tm Rinst e3 e2, ex ``(congr_lam' %%e4 %%e5)
  
  else if (is_ren_tm e1)
  then do (e3,e4) ← destruct_ren_tm e1, e5 ← normalize_tm Comp e4 e1, e6 ← normalize_tm Rinst e3 e5, ex ``(compRenRen'_tm %%e5 %%e2 %%e6)
  
  else ex ``(extRen'_tm %%e1 %%e2)
  
  else if (n = Comp)
  then if (is_scons e1)
  then do (e3,e4) ← destruct_scons e1, e5 ← normalize_tm Inst e3 e2, e6 ← normalize_tm Comp e4 e2, e7 ← normalize_tm Cons e5 e6, ex ``(scons_comp' %%e5 %%e1 %%e6)
  
  else if (is_subst_tm e1 && is_var_tm e2)
  then do (e3,e4) ← destruct_subst_tm e1, ex ``(varL_tm %%e3)
  
  else ex ``(comp' %%e1 %%e2)
  
  else ex ``(%%e1=%%e2)
